"""
Author: Remi Lafage <remi.lafage@onera.fr>

This package is distributed under New BSD license.

Norm function.
"""

import numpy as np

from smt.problems.problem import Problem


class LpNorm(Problem):
    def _initialize(self, ndim=1):
        self.options.declare("order", default=2, types=int)
        self.options.declare("name", "LpNorm", types=str)

    def _setup(self):
        self.xlimits[:, 0] = -1.0
        self.xlimits[:, 1] = 1.0

    def _evaluate(self, x, kx):
        """
        Arguments
        ---------
        x : ndarray[ne, ndim]
            Evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[ne, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """

        ne, nx = x.shape
        p = self.options["order"]
        assert p > 0

        y = np.zeros((ne, 1), complex)
        lp_norm = np.sum(np.abs(x) ** p, axis=-1) ** (1.0 / p)
        if kx is None:
            y[:, 0] = lp_norm
        else:
            y[:, 0] = np.sign(x[:, kx]) * (np.absolute(x[:, kx]) / lp_norm) ** (p - 1)

        return y



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

Base class for sampling algorithms.
"""

from abc import ABCMeta, abstractmethod

import numpy as np

from smt.utils.options_dictionary import OptionsDictionary


class SamplingMethod(metaclass=ABCMeta):
    def __init__(self, **kwargs):
        """
        Constructor where values of options can be passed in.
        xlimits keyword argument is required.

        For the list of options, see the documentation for the sampling method being used.

        Parameters
        ----------
        **kwargs : named arguments
            Set of options that can be optionally set; each option must have been declared.

        Examples
        --------
        >>> import numpy as np
        >>> from smt.sampling_methods import Random
        >>> sampling = Random(xlimits=np.arange(2).reshape((1, 2)))
        """
        self.options = OptionsDictionary()
        self.options.declare(
            "xlimits",
            types=np.ndarray,
            desc="The interval of the domain in each dimension with shape nx x 2 (required)",
        )
        self._initialize(**kwargs)
        self.options.update(kwargs)
        if self.options["xlimits"] is None:
            raise ValueError("xlimits keyword argument is required")

    def _initialize(self, **kwargs) -> None:
        """
        Implemented by sampling methods to declare options
        and/or use these optional values for initialization (optional)

        Parameters
        ----------
        **kwargs : named arguments passed by the user
            Set of options that can be optionally set

        Examples
        --------
        self.options.declare('option_name', default_value, types=(bool, int), desc='description')
        """
        pass

    def __call__(self, nt: int = None) -> np.ndarray:
        """
        Compute the samples.
        Depending on the concrete sampling method the requested number of samples nt may not be enforced.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Parameters
        ----------
        nt : int
            Number of points hint.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the input space.
        """
        return self._compute(nt)

    @abstractmethod
    def _compute(self, nt: int = None) -> np.ndarray:
        """
        Implemented by sampling methods to compute the samples.
        Depending on the concrete sampling method the requested number of samples nt may not be enforced.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Parameters
        ----------
        nt : int
            Number of points requested.
            Depending on the concrete sampling method this requested number of samples may not be enforced.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the input space.
        """
        raise Exception("This sampling method has not been implemented correctly")


class ScaledSamplingMethod(SamplingMethod):
    """This class represents sample methods which generates samples in the unit hypercube [0, 1]^nx.

    The __call__ method does scale the generated samples accordingly to the defined xlimits.

    Implementation notes:

    * When nt is None, it defaults to 2 * nx.
    * xlimits is presence is checked. ValueError is raised if not specified.
    """

    def __call__(self, nt: int = None) -> np.ndarray:
        """
        Compute the samples.
        Depending on the concrete sampling method the requested number of samples nt may not be enforced.
        When nt is None, it defaults to 2 * nx.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Parameters
        ----------
        nt : int (optional, default 2*nx)
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the input space.
        """
        xlimits = self.options["xlimits"]
        if nt is None:
            nt = 2 * xlimits.shape[0]
        return _scale_to_xlimits(self._compute(nt), xlimits)

    @abstractmethod
    def _compute(self, nt: int = None) -> np.ndarray:
        """
        Implemented by sampling methods to compute the requested number of sampling points.

        The number of dimensions (nx) is determined based on `xlimits.shape[0]`.

        Parameters
        ----------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the unit hypercube.
        """
        raise Exception("This sampling method has not been implemented correctly")


def _scale_to_xlimits(samples: np.ndarray, xlimits: np.ndarray) -> np.ndarray:
    """Scales the samples from the unit hypercube to the specified limits.

    Parameters
    ----------
    samples : np.ndarray
        The samples with coordinates in [0,1]
    xlimits : np.ndarray
        The xlimits

    Returns
    -------
    np.ndarray
        The scaled samples.
    """
    nx = xlimits.shape[0]
    for kx in range(nx):
        samples[:, kx] = xlimits[kx, 0] + samples[:, kx] * (
            xlimits[kx, 1] - xlimits[kx, 0]
        )
    return samples



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.
"""

import contextlib

import numpy as np
import scipy.linalg
import scipy.sparse.linalg

from smt.utils.options_dictionary import OptionsDictionary

VALID_SOLVERS = (
    "krylov-dense",
    "dense-lu",
    "dense-chol",
    "lu",
    "ilu",
    "krylov",
    "krylov-lu",
    "krylov-mg",
    "gs",
    "jacobi",
    "mg",
    "null",
)


def get_solver(solver):
    if solver == "dense-lu":
        return DenseLUSolver()
    elif solver == "dense-chol":
        return DenseCholeskySolver()
    elif solver == "krylov-dense":
        return KrylovSolver(pc="dense")
    elif solver == "lu" or solver == "ilu":
        return DirectSolver(alg=solver)
    elif solver == "krylov":
        return KrylovSolver()
    elif solver == "krylov-lu":
        return KrylovSolver(pc="lu")
    elif solver == "krylov-mg":
        return KrylovSolver(pc="mg")
    elif solver == "gs" or solver == "jacobi":
        return StationarySolver(solver=solver)
    elif solver == "mg":
        return MultigridSolver()
    elif isinstance(solver, LinearSolver):
        return solver
    elif solver == "null":
        return NullSolver()
    elif solver is None:
        return None


class Callback(object):
    def __init__(self, size, string, interval, printer):
        self.size = size
        self.string = string
        self.interval = interval
        self.printer = printer

        self.counter = 0
        self.ind_y = 0
        self.mtx = None
        self.rhs = None
        self.norm0 = 1.0

    def _print_norm(self, norm):
        if self.counter == 0:
            self.norm0 = norm

        if self.counter % self.interval == 0:
            self.printer(
                "%s (%i x %i mtx), output %-3i : %3i  %15.9e  %15.9e"
                % (
                    self.string,
                    self.size,
                    self.size,
                    self.ind_y,
                    self.counter,
                    norm,
                    norm / self.norm0,
                )
            )
        self.counter += 1

    def _print_res(self, res):
        self._print_norm(res)

    def _print_sol(self, sol):
        res = self.mtx.dot(sol) - self.rhs
        norm = np.linalg.norm(res)
        self._print_norm(norm)


class LinearSolver(object):
    def __init__(self, **kwargs):
        self.mtx = None
        self.rhs = None

        self.options = OptionsDictionary()
        self.options.declare("print_init", True, types=bool)
        self.options.declare("print_solve", True, types=bool)
        self._initialize()
        self.options.update(kwargs)

    def _initialize(self):
        pass

    def _setup(self, mtx, printer, mg_matrices=[]):
        pass

    def _solve(self, rhs, sol=None, ind_y=0):
        pass

    def _clone(self):
        clone = self.__class__()
        clone.options.update(clone.options._dict)
        return clone

    @contextlib.contextmanager
    def _active(self, active):
        orig_active = self.printer.active

        self.printer.active = self.printer.active and active
        yield self.printer
        self.printer.active = orig_active


class NullSolver(LinearSolver):
    def solve(self, rhs, sol=None, ind_y=0):
        pass


class DenseCholeskySolver(LinearSolver):
    def _setup(self, mtx, printer, mg_matrices=[]):
        self.printer = printer
        with self._active(self.options["print_init"]) as printer:
            self.mtx = mtx.A
            assert isinstance(self.mtx, np.ndarray), "mtx is of type %s" % type(mtx)

            with printer._timed_context(
                "Performing Chol. fact. (%i x %i mtx)" % mtx.shape
            ):
                self.upper = scipy.linalg.cholesky(self.mtx)

    def _solve(self, rhs, sol=None, ind_y=0):
        with self._active(self.options["print_solve"]) as printer:
            self.rhs = rhs

            if sol is None:
                sol = np.array(rhs)

            with printer._timed_context("Back solving (%i x %i mtx)" % self.mtx.shape):
                sol[:] = rhs
                scipy.linalg.solve_triangular(
                    self.upper, sol, overwrite_b=True, trans="T"
                )
                scipy.linalg.solve_triangular(self.upper, sol, overwrite_b=True)

        return sol


class DenseLUSolver(LinearSolver):
    def _setup(self, mtx, printer, mg_matrices=[]):
        self.printer = printer
        with self._active(self.options["print_init"]) as printer:
            self.mtx = mtx
            assert isinstance(mtx, np.ndarray), "mtx is of type %s" % type(mtx)

            with printer._timed_context(
                "Performing LU fact. (%i x %i mtx)" % mtx.shape
            ):
                self.fact = scipy.linalg.lu_factor(mtx)

    def _solve(self, rhs, sol=None, ind_y=0):
        with self._active(self.options["print_solve"]) as printer:
            self.rhs = rhs

            if sol is None:
                sol = np.array(rhs)

            with printer._timed_context("Back solving (%i x %i mtx)" % self.mtx.shape):
                sol[:] = scipy.linalg.lu_solve(self.fact, rhs)

        return sol


class DirectSolver(LinearSolver):
    def _initialize(self):
        self.options.declare("alg", "lu", values=["lu", "ilu"])

    def _setup(self, mtx, printer, mg_matrices=[]):
        self.printer = printer
        with self._active(self.options["print_init"]) as printer:
            self.mtx = mtx
            assert isinstance(mtx, scipy.sparse.spmatrix), "mtx is of type %s" % type(
                mtx
            )

            with printer._timed_context(
                "Performing %s fact. (%i x %i mtx)"
                % ((self.options["alg"],) + mtx.shape)
            ):
                if self.options["alg"] == "lu":
                    self.fact = scipy.sparse.linalg.splu(mtx)
                elif self.options["alg"] == "ilu":
                    self.fact = scipy.sparse.linalg.spilu(
                        mtx, drop_rule="interp", drop_tol=1e-3, fill_factor=2
                    )

    def _solve(self, rhs, sol=None, ind_y=0):
        with self._active(self.options["print_solve"]) as printer:
            self.rhs = rhs

            if sol is None:
                sol = np.array(rhs)

            with printer._timed_context("Back solving (%i x %i mtx)" % self.mtx.shape):
                sol[:] = self.fact.solve(rhs)

        return sol


class KrylovSolver(LinearSolver):
    def _initialize(self):
        self.options.declare("interval", 10, types=int)
        self.options.declare("solver", "cg", values=["cg", "bicgstab", "gmres"])
        self.options.declare(
            "pc",
            None,
            values=[None, "ilu", "lu", "gs", "jacobi", "mg", "dense"],
            types=LinearSolver,
        )
        self.options.declare("ilimit", 100, types=int)
        self.options.declare("atol", 1e-15, types=(int, float))
        self.options.declare("rtol", 1e-15, types=(int, float))

    def _setup(self, mtx, printer, mg_matrices=[]):
        self.printer = printer
        with self._active(self.options["print_init"]) as printer:
            self.mtx = mtx

            pc_solver = get_solver(self.options["pc"])

            if pc_solver is not None:
                pc_solver._setup(mtx, printer, mg_matrices=mg_matrices)
                self.pc_solver = pc_solver
                self.pc_op = scipy.sparse.linalg.LinearOperator(
                    mtx.shape, matvec=pc_solver._solve
                )
            else:
                self.pc_solver = None
                self.pc_op = None

            self.callback = Callback(
                mtx.shape[0], "Krylov solver", self.options["interval"], printer
            )
            if self.options["solver"] == "cg":
                self.solver = scipy.sparse.linalg.cg
                self.callback_func = self.callback._print_sol
                self.solver_kwargs = {
                    "atol": 0.0,
                    "rtol": self.options["rtol"],
                    "maxiter": self.options["ilimit"],
                }
            elif self.options["solver"] == "bicgstab":
                self.solver = scipy.sparse.linalg.bicgstab
                self.callback_func = self.callback._print_sol
                self.solver_kwargs = {
                    "rtol": self.options["rtol"],
                    "maxiter": self.options["ilimit"],
                }
            elif self.options["solver"] == "gmres":
                self.solver = scipy.sparse.linalg.gmres
                self.callback_func = self.callback._print_res
                self.solver_kwargs = {
                    "rtol": self.options["rtol"],
                    "maxiter": self.options["ilimit"],
                    "restart": min(self.options["ilimit"], mtx.shape[0]),
                }
            self._patch_when_scipy_lessthan_v111()

    def _patch_when_scipy_lessthan_v111(self):
        """
        From scipy 1.11.0 release notes
        The tol argument of scipy.sparse.linalg.{bcg,bicstab,cg,cgs,gcrotmk,gmres,lgmres,minres,qmr,tfqmr}
        is now deprecated in favour of rtol and will be removed in SciPy 1.14.
        Furthermore, the default value of atol for these functions is due to change to 0.0 in SciPy 1.14.
        """
        import scipy

        scipy_version = scipy.__version__
        version_tuple = tuple(map(int, scipy_version.split(".")))
        is_greater_than_1_11 = version_tuple[0] > 1 or (
            version_tuple[0] == 1 and version_tuple[1] >= 11
        )

        if not is_greater_than_1_11:
            self.solver_kwargs["tol"] = self.solver_kwargs["rtol"]
            del self.solver_kwargs["rtol"]

    def _solve(self, rhs, sol=None, ind_y=0):
        with self._active(self.options["print_solve"]) as printer:
            self.rhs = rhs

            if sol is None:
                sol = np.array(rhs)

            with printer._timed_context(
                "Running %s Krylov solver (%i x %i mtx)"
                % ((self.options["solver"],) + self.mtx.shape)
            ):
                self.callback.counter = 0
                self.callback.ind_y = ind_y
                self.callback.mtx = self.mtx
                self.callback.rhs = rhs

                self.callback._print_sol(sol)

                tmp, _ = self.solver(
                    self.mtx,
                    rhs,
                    x0=sol,
                    M=self.pc_op,
                    callback=self.callback_func,
                    **self.solver_kwargs,
                )

            sol[:] = tmp

        return sol


class StationarySolver(LinearSolver):
    def _initialize(self):
        self.options.declare("interval", 10, types=int)
        self.options.declare("solver", "gs", values=["gs", "jacobi"])
        self.options.declare("damping", 1.0, types=(int, float))
        self.options.declare("ilimit", 10, types=int)

    def _setup(self, mtx, printer, mg_matrices=[]):
        self.printer = printer
        with self._active(self.options["print_init"]) as printer:
            self.mtx = mtx

            self.callback = Callback(
                mtx.shape[0], "Stationary solver", self.options["interval"], printer
            )

            with printer._timed_context(
                "Initializing %s solver (%i x %i mtx)"
                % ((self.options["solver"],) + self.mtx.shape)
            ):
                if self.options["solver"] == "jacobi":
                    # A x = b
                    # x_{k+1} = x_k + w D^{-1} (b - A x_k)
                    self.d_inv = self.options["damping"] / self._split_mtx_diag()
                    self.iterate = self._jacobi

                elif self.options["solver"] == "gs":
                    # A x = b
                    # x_{k+1} = x_k + (1/w D + L)^{-1} (b - A x_k)
                    mtx_d = self._split_mtx("diag")
                    mtx_l = self._split_mtx("lower")
                    mtx_ldw = mtx_l + mtx_d / self.options["damping"]
                    self.inv = scipy.sparse.linalg.splu(mtx_ldw)
                    self.iterate = self._gs

    def _split_mtx_diag(self):
        shape = self.mtx.shape
        rows, cols, data = scipy.sparse.find(self.mtx)

        mask_d = rows == cols
        diag = np.zeros(shape[0])
        np.add.at(diag, rows[mask_d], data[mask_d])
        return diag

    def _split_mtx(self, part):
        shape = self.mtx.shape
        rows, cols, data = scipy.sparse.find(self.mtx)

        if part == "diag":
            mask = rows == cols
        elif part == "lower":
            mask = rows > cols
        elif part == "upper":
            mask = rows < cols

        return scipy.sparse.csc_matrix(
            (data[mask], (rows[mask], cols[mask])), shape=shape
        )

    def _jacobi(self, rhs, sol):
        # A x = b
        # x_{k+1} = x_k + w D^{-1} (b - A x_k)
        sol += self.d_inv * (rhs - self.mtx.dot(sol))

    def _gs(self, rhs, sol):
        # A x = b
        # x_{k+1} = x_k + (1/w D + L)^{-1} (b - A x_k)
        sol += self.inv.solve(rhs - self.mtx.dot(sol))

    def _solve(self, rhs, sol=None, ind_y=0):
        with self._active(self.options["print_solve"]) as printer:
            self.rhs = rhs

            if sol is None:
                sol = np.array(rhs)

            self.callback.counter = 0
            self.callback.ind_y = ind_y
            self.callback.mtx = self.mtx
            self.callback.rhs = rhs

            with printer._timed_context(
                "Running %s stationary solver (%i x %i mtx)"
                % ((self.options["solver"],) + self.mtx.shape)
            ):
                for ind in range(self.options["ilimit"]):
                    self.iterate(rhs, sol)
                    self.callback._print_sol(sol)

        return sol


class MultigridSolver(LinearSolver):
    def _initialize(self):
        self.options.declare("interval", 1, types=int)
        self.options.declare("mg_cycles", 0, types=int)
        self.options.declare(
            "solver",
            "null",
            values=["null", "gs", "jacobi", "krylov"],
            types=LinearSolver,
        )

    def _setup(self, mtx, printer, mg_matrices=[]):
        self.printer = printer
        with self._active(self.options["print_init"]) as printer:
            self.mtx = mtx

            solver = get_solver(self.options["solver"])
            mg_solver = solver._clone()
            mg_solver._setup(mtx, printer)

            self.mg_mtx = [mtx]
            self.mg_sol = [np.zeros(self.mtx.shape[0])]
            self.mg_rhs = [np.zeros(self.mtx.shape[0])]
            self.mg_ops = []
            self.mg_solvers = [mg_solver]

            for ind, mg_op in enumerate(mg_matrices):
                mg_mtx = mg_op.T.dot(self.mg_mtx[-1]).dot(mg_op).tocsc()
                mg_sol = mg_op.T.dot(self.mg_sol[-1])
                mg_rhs = mg_op.T.dot(self.mg_rhs[-1])

                mg_solver = solver._clone()
                mg_solver._setup(mg_mtx, printer)

                self.mg_mtx.append(mg_mtx)
                self.mg_sol.append(mg_sol)
                self.mg_rhs.append(mg_rhs)
                self.mg_ops.append(mg_op)
                self.mg_solvers.append(mg_solver)

            mg_mtx = self.mg_mtx[-1]
            mg_solver = DirectSolver()
            mg_solver._setup(mg_mtx, printer)
            self.mg_solvers[-1] = mg_solver

            self.callback = Callback(
                mtx.shape[0], "Multigrid solver", self.options["interval"], printer
            )

    def _restrict(self, ind_level):
        mg_op = self.mg_ops[ind_level]
        mtx = self.mg_mtx[ind_level]
        sol = self.mg_sol[ind_level]
        rhs = self.mg_rhs[ind_level]

        res = rhs - mtx.dot(sol)
        res_coarse = mg_op.T.dot(res)
        self.mg_rhs[ind_level + 1][:] = res_coarse

    def _smooth_and_restrict(self, ind_level, ind_cycle, ind_y):
        mg_op = self.mg_ops[ind_level]
        mtx = self.mg_mtx[ind_level]
        sol = self.mg_sol[ind_level]
        rhs = self.mg_rhs[ind_level]
        solver = self.mg_solvers[ind_level]
        solver.print_info = "MG iter %i level %i" % (ind_cycle, ind_level)

        solver._solve(rhs, sol, ind_y)

        res = rhs - mtx.dot(sol)
        res_coarse = mg_op.T.dot(res)
        self.mg_rhs[ind_level + 1][:] = res_coarse

    def _coarse_solve(self, ind_cycle, ind_y):
        sol = self.mg_sol[-1]
        rhs = self.mg_rhs[-1]
        solver = self.mg_solvers[-1]
        solver.print_info = "MG iter %i level %i" % (ind_cycle, len(self.mg_ops))
        solver._solve(rhs, sol, ind_y)

    def _smooth_and_interpolate(self, ind_level, ind_cycle, ind_y):
        mg_op = self.mg_ops[ind_level]
        _mtx = self.mg_mtx[ind_level]
        sol = self.mg_sol[ind_level]
        rhs = self.mg_rhs[ind_level]
        solver = self.mg_solvers[ind_level]
        solver.print_info = "MG iter %i level %i" % (ind_cycle, ind_level)

        sol_coarse = self.mg_sol[ind_level + 1]
        sol += mg_op.dot(sol_coarse)

        solver._solve(rhs, sol, ind_y)

    def _solve(self, rhs, sol=None, ind_y=0):
        with self._active(self.options["print_solve"]) as _printer:
            self.rhs = rhs

            if sol is None:
                sol = np.array(rhs)

            orig_sol = sol

            self.callback.counter = 0
            self.callback.ind_y = ind_y
            self.callback.mtx = self.mtx
            self.callback.rhs = rhs

            self.mg_rhs[0][:] = rhs

            for ind_level in range(len(self.mg_ops)):
                self._restrict(ind_level)

            self._coarse_solve(-1, ind_y)

            for ind_level in range(len(self.mg_ops) - 1, -1, -1):
                self._smooth_and_interpolate(ind_level, -1, ind_y)

            for ind_cycle in range(self.options["mg_cycles"]):
                for ind_level in range(len(self.mg_ops)):
                    self._smooth_and_restrict(ind_level, ind_cycle, ind_y)

                self._coarse_solve(ind_cycle, ind_y)

                for ind_level in range(len(self.mg_ops) - 1, -1, -1):
                    self._smooth_and_interpolate(ind_level, ind_cycle, ind_y)

            orig_sol[:] = self.mg_sol[0]

        return orig_sol



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>
        Dr. Nathalie.bartoli      <nathalie@onera.fr>

This package is distributed under New BSD license.

TO DO:
- define outputs['sol'] = self.sol
"""

import numpy as np
from sklearn import linear_model

from smt.surrogate_models.surrogate_model import SurrogateModel
from smt.utils.caching import cached_operation


class LS(SurrogateModel):
    """
    Least square model.
    This model uses the linear_model.LinearRegression class from scikit-learn.
    Default-parameters from scikit-learn are used herein.
    """

    name = "LS"

    def _initialize(self):
        super(LS, self)._initialize()
        declare = self.options.declare
        supports = self.supports
        declare(
            "data_dir",
            values=None,
            types=str,
            desc="Directory for loading / saving cached data; None means do not save or load",
        )

        supports["derivatives"] = True

    ############################################################################
    # Model functions
    ############################################################################

    def _new_train(self):
        """
        Train the model
        """
        X = self.training_points[None][0][0]
        y = self.training_points[None][0][1]
        self.mod = linear_model.LinearRegression()
        self.mod.fit(X, y)

    def _train(self):
        """
        Train the model
        """
        inputs = {"self": self}
        with cached_operation(inputs, self.options["data_dir"]) as outputs:
            if outputs:
                self.sol = outputs["sol"]
            else:
                self._new_train()

    def _predict_values(self, x):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        y = self.mod.predict(x)
        return y

    def _predict_derivatives(self, x, kx):
        """
        Evaluates the derivatives at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        kx : int
            The 0-based index of the input variable with respect to which derivatives are desired.

        Returns
        -------
        y : np.ndarray
            Derivative values.
        """

        # Initialization
        n_eval, n_features_x = x.shape
        y = np.ones((n_eval, self.ny)) * self.mod.coef_[:, kx]
        return y



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

Cantilever beam problem from:

Liu, H., Xu, S., & Wang, X. Sampling strategies and metamodeling techniques for engineering design:
comparison and application. In ASME Turbo Expo 2016: Turbomachinery Technical Conference and Exposition.
American Society of Mechanical Engineers. June, 2016.

Cheng, G. H., Younis, A., Hajikolaei, K. H., and Wang, G. G.
Trust Region Based Mode Pursuing Sampling Method for Global Optimization of High Dimensional Design Problems.
Journal of Mechanical Design, 137(2). 2015.
"""

import numpy as np

from smt.problems.problem import Problem


class CantileverBeam(Problem):
    def _initialize(self):
        self.options.declare("name", "CantileverBeam", types=str)
        self.options.declare("ndim", 3, types=int)
        self.options.declare("P", 50e3, types=(int, float), desc="Tip load (50 kN)")
        self.options.declare(
            "E", 200e9, types=(int, float), desc="Modulus of elast. (200 GPa)"
        )

    def _setup(self):
        assert self.options["ndim"] % 3 == 0, "ndim must be divisible by 3"

        # Width b
        self.xlimits[0::3, 0] = 0.01
        self.xlimits[0::3, 1] = 0.05

        # Height h
        self.xlimits[1::3, 0] = 0.30
        self.xlimits[1::3, 1] = 0.65

        # Length l
        self.xlimits[2::3, 0] = 0.5
        self.xlimits[2::3, 1] = 1.0

    def _evaluate(self, x, kx):
        """
        Arguments
        ---------
        x : ndarray[ne, nx]
            Evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[ne, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """
        ne, nx = x.shape

        nelem = int(self.options["ndim"] / 3)
        P = self.options["P"]
        E = self.options["E"]

        y = np.zeros((ne, 1), complex)
        if kx is None:
            for ielem in range(nelem):
                b = x[:, 3 * ielem + 0]
                h = x[:, 3 * ielem + 1]

                y[:, 0] += (
                    12.0 / b / h**3 * np.sum(x[:, 2 + 3 * ielem :: 3], axis=1) ** 3
                )
                y[:, 0] -= (
                    12.0 / b / h**3 * np.sum(x[:, 5 + 3 * ielem :: 3], axis=1) ** 3
                )
        else:
            kelem = int(np.floor(kx / 3))
            if kx % 3 == 0:
                b = x[:, 3 * kelem + 0]
                h = x[:, 3 * kelem + 1]
                y[:, 0] += (
                    -12.0 / b**2 / h**3 * np.sum(x[:, 2 + 3 * kelem :: 3], axis=1) ** 3
                )
                y[:, 0] -= (
                    -12.0 / b**2 / h**3 * np.sum(x[:, 5 + 3 * kelem :: 3], axis=1) ** 3
                )

            elif kx % 3 == 1:
                b = x[:, 3 * kelem + 0]
                h = x[:, 3 * kelem + 1]
                y[:, 0] += (
                    -36.0 / b / h**4 * np.sum(x[:, 2 + 3 * kelem :: 3], axis=1) ** 3
                )
                y[:, 0] -= (
                    -36.0 / b / h**4 * np.sum(x[:, 5 + 3 * kelem :: 3], axis=1) ** 3
                )
            elif kx % 3 == 2:
                for ielem in range(kelem + 1):
                    b = x[:, 3 * ielem + 0]
                    h = x[:, 3 * ielem + 1]
                    y[:, 0] += (
                        36.0 / b / h**3 * np.sum(x[:, 2 + 3 * ielem :: 3], axis=1) ** 2
                    )
                    if kelem > ielem:
                        y[:, 0] -= (
                            36.0
                            / b
                            / h**3
                            * np.sum(x[:, 5 + 3 * ielem :: 3], axis=1) ** 2
                        )

        return (P / 3 / E) * y



from .full_factorial import FullFactorial
from .lhs import LHS
from .pydoe import BoxBehnken, Factorial, Gsd, PlackettBurman
from .random import Random

__all__ = [
    "Random",
    "LHS",
    "FullFactorial",
    "BoxBehnken",
    "PlackettBurman",
    "Factorial",
    "Gsd",
]



from .branin import Branin
from .cantilever_beam import CantileverBeam
from .hierarchical_goldstein import HierarchicalGoldstein
from .lp_norm import LpNorm
from .mixed_cantilever_beam import MixedCantileverBeam
from .ndim_cantilever_beam import NdimCantileverBeam
from .ndim_robot_arm import NdimRobotArm
from .ndim_rosenbrock import NdimRosenbrock
from .ndim_step_function import NdimStepFunction
from .neural_network import HierarchicalNeuralNetwork
from .reduced_problem import ReducedProblem
from .robot_arm import RobotArm
from .rosenbrock import Rosenbrock
from .sphere import Sphere
from .tensor_product import TensorProduct
from .torsion_vibration import TorsionVibration
from .water_flow import WaterFlow
from .water_flow_lfidelity import WaterFlowLFidelity
from .welded_beam import WeldedBeam
from .wing_weight import WingWeight

__all__ = [
    "Branin",
    "CantileverBeam",
    "Sphere",
    "ReducedProblem",
    "RobotArm",
    "Rosenbrock",
    "TensorProduct",
    "TorsionVibration",
    "WaterFlow",
    "WaterFlowLFidelity",
    "WeldedBeam",
    "WingWeight",
    "NdimCantileverBeam",
    "MixedCantileverBeam",
    "HierarchicalNeuralNetwork",
    "HierarchicalGoldstein",
    "NdimRobotArm",
    "NdimRosenbrock",
    "NdimStepFunction",
    "LpNorm",
]



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

N-dimensional step function problem.
"""

from smt.problems.problem import Problem
from smt.problems.tensor_product import TensorProduct
from smt.utils.options_dictionary import OptionsDictionary


class NdimStepFunction(Problem):
    def __init__(self, ndim=1, width=10.0):
        super().__init__()
        self.problem = TensorProduct(ndim=ndim, func="tanh", width=width)

        self.options = OptionsDictionary()
        self.options.declare("ndim", ndim, types=int)
        self.options.declare("return_complex", False, types=bool)
        self.options.declare("name", "NdimStepFunction", types=str)

        self.xlimits = self.problem.xlimits

    def _evaluate(self, x, kx):
        return self.problem._evaluate(x, kx)



"""
Author: Jasper Bussemaker <jasper.bussemaker@dlr.de>

This package is distributed under New BSD license.
"""

from typing import List, Optional, Sequence, Tuple, Union

import numpy as np

from smt.sampling_methods import LHS

try:
    from ConfigSpace import (
        CategoricalHyperparameter,
        Configuration,
        ConfigurationSpace,
        EqualsCondition,
        ForbiddenAndConjunction,
        ForbiddenEqualsClause,
        ForbiddenInClause,
        InCondition,
        OrdinalHyperparameter,
        UniformFloatHyperparameter,
        UniformIntegerHyperparameter,
        ForbiddenLessThanRelation,
    )
    from ConfigSpace.exceptions import ForbiddenValueError
    from ConfigSpace.util import get_random_neighbor

    HAS_CONFIG_SPACE = True

except ImportError:
    HAS_CONFIG_SPACE = False

    class Configuration:
        pass

    class ConfigurationSpace:
        pass

    class UniformIntegerHyperparameter:
        pass


def ensure_design_space(xt=None, xlimits=None, design_space=None) -> "BaseDesignSpace":
    """Interface to turn legacy input formats into a DesignSpace"""

    if design_space is not None and isinstance(design_space, BaseDesignSpace):
        return design_space

    if xlimits is not None:
        return DesignSpace(xlimits)

    if xt is not None:
        return DesignSpace([[np.min(xt) - 0.99, np.max(xt) + 1e-4]] * xt.shape[1])

    raise ValueError("Nothing defined that could be interpreted as a design space!")


class DesignVariable:
    """Base class for defining a design variable"""

    upper: Union[float, int]
    lower: Union[float, int]

    def get_typename(self):
        return self.__class__.__name__

    def get_limits(self) -> Union[list, tuple]:
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError


class FloatVariable(DesignVariable):
    """A continuous design variable, varying between its lower and upper bounds"""

    def __init__(self, lower: float, upper: float):
        if upper <= lower:
            raise ValueError(
                f"Upper bound should be higher than lower bound: {upper} <= {lower}"
            )
        self.lower = lower
        self.upper = upper

    def get_limits(self) -> Tuple[float, float]:
        return self.lower, self.upper

    def __str__(self):
        return f"Float ({self.lower}, {self.upper})"

    def __repr__(self):
        return f"{self.get_typename()}({self.lower}, {self.upper})"


class IntegerVariable(DesignVariable):
    """An integer variable that can take any integer value between the bounds (inclusive)"""

    def __init__(self, lower: int, upper: int):
        if upper <= lower:
            raise ValueError(
                f"Upper bound should be higher than lower bound: {upper} <= {lower}"
            )
        self.lower = lower
        self.upper = upper

    def get_limits(self) -> Tuple[int, int]:
        return self.lower, self.upper

    def __str__(self):
        return f"Int ({self.lower}, {self.upper})"

    def __repr__(self):
        return f"{self.get_typename()}({self.lower}, {self.upper})"


class OrdinalVariable(DesignVariable):
    """An ordinal variable that can take any of the given value, and where order between the values matters"""

    def __init__(self, values: List[Union[str, int, float]]):
        if len(values) < 2:
            raise ValueError(f"There should at least be 2 values: {values}")
        self.values = values

    @property
    def lower(self) -> int:
        return 0

    @property
    def upper(self) -> int:
        return len(self.values) - 1

    def get_limits(self) -> List[str]:
        # We convert to integer strings for compatibility reasons
        return [str(i) for i in range(len(self.values))]

    def __str__(self):
        return f"Ord {self.values}"

    def __repr__(self):
        return f"{self.get_typename()}({self.values})"


class CategoricalVariable(DesignVariable):
    """A categorical variable that can take any of the given values, and where order does not matter"""

    def __init__(self, values: List[Union[str, int, float]]):
        if len(values) < 2:
            raise ValueError(f"There should at least be 2 values: {values}")
        self.values = values

    @property
    def lower(self) -> int:
        return 0

    @property
    def upper(self) -> int:
        return len(self.values) - 1

    @property
    def n_values(self):
        return len(self.values)

    def get_limits(self) -> List[Union[str, int, float]]:
        # We convert to strings for compatibility reasons
        return [str(value) for value in self.values]

    def __str__(self):
        return f"Cat {self.values}"

    def __repr__(self):
        return f"{self.get_typename()}({self.values})"


class BaseDesignSpace:
    """
    Interface for specifying (hierarchical) design spaces.

    This class itself only specifies the functionality that any design space definition should implement:
    - a way to specify the design variables, their types, and their bounds or options
    - a way to correct a set of design vectors such that they satisfy all design space hierarchy constraints
    - a way to query which design variables are acting for a set of design vectors
    - a way to impute a set of design vectors such that non-acting design variables are assigned some default value
    - a way to sample n valid design vectors from the design space

    If you want to actually define a design space, use the `DesignSpace` class!

    Note that the correction, querying, and imputation mechanisms should all be implemented in one function
    (`correct_get_acting`), as usually these operations are tightly related.
    """

    def __init__(self, design_variables: List[DesignVariable] = None):
        self._design_variables = design_variables
        self._is_cat_mask = None
        self._is_conditionally_acting_mask = None
        self.seed = None
        self.has_valcons_ord_int = False

    @property
    def design_variables(self) -> List[DesignVariable]:
        if self._design_variables is None:
            self._design_variables = dvs = self._get_design_variables()
            if dvs is None:
                raise RuntimeError(
                    "Design space should either specify the design variables upon initialization "
                    "or as output from _get_design_variables!"
                )
        return self._design_variables

    @property
    def is_cat_mask(self) -> np.ndarray:
        """Boolean mask specifying for each design variable whether it is a categorical variable"""
        if self._is_cat_mask is None:
            self._is_cat_mask = np.array(
                [isinstance(dv, CategoricalVariable) for dv in self.design_variables]
            )
        return self._is_cat_mask

    @property
    def is_all_cont(self) -> bool:
        """Whether or not the space is continuous"""
        is_continuous = all(
            isinstance(dv, FloatVariable) for dv in self.design_variables
        )
        return is_continuous

    @property
    def is_conditionally_acting(self) -> np.ndarray:
        """Boolean mask specifying for each design variable whether it is conditionally acting (can be non-acting)"""
        if self._is_conditionally_acting_mask is None:
            self._is_conditionally_acting_mask = self._is_conditionally_acting()
        return self._is_conditionally_acting_mask

    @property
    def n_dv(self) -> int:
        """Get the number of design variables"""
        return len(self.design_variables)

    def correct_get_acting(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Correct the given matrix of design vectors and return the corrected vectors and the is_acting matrix.
        It is automatically detected whether input is provided in unfolded space or not.

        Parameters
        ----------
        x: np.ndarray [n_obs, dim]
           - Input variables

        Returns
        -------
        x_corrected: np.ndarray [n_obs, dim]
           - Corrected and imputed input variables
        is_acting: np.ndarray [n_obs, dim]
           - Boolean matrix specifying for each variable whether it is acting or non-acting
        """

        # Detect whether input is provided in unfolded space
        x = np.atleast_2d(x)
        if x.shape[1] == self.n_dv:
            x_is_unfolded = False
        elif x.shape[1] == self._get_n_dim_unfolded():
            x_is_unfolded = True
        else:
            raise ValueError(f"Incorrect shape, expecting {self.n_dv} columns!")

        # If needed, fold before correcting
        if x_is_unfolded:
            x, _ = self.fold_x(x)

        indi = 0
        for i in self.design_variables:
            if not (isinstance(i, FloatVariable)):
                x[:, indi] = np.int64(np.round(x[:, indi], 0))
            indi += 1

        # Correct and get the is_acting matrix
        x_corrected, is_acting = self._correct_get_acting(x)

        # Check conditionally-acting status
        if np.any(~is_acting[:, ~self.is_conditionally_acting]):
            raise RuntimeError("Unconditionally acting variables cannot be non-acting!")

        # Unfold if needed
        if x_is_unfolded:
            x_corrected, is_acting = self.unfold_x(x_corrected, is_acting)

        return x_corrected, is_acting

    def decode_values(
        self, x: np.ndarray, i_dv: int = None
    ) -> List[Union[str, int, float, list]]:
        """
        Return decoded values: converts ordinal and categorical back to their original values.

        If i_dv is given, decoding is done for one specific design variable only.
        If i_dv=None, decoding will be done for all design variables: 1d input is interpreted as a design vector,
        2d input is interpreted as a set of design vectors.
        """

        def _decode_dv(x_encoded: np.ndarray, i_dv_decode):
            dv = self.design_variables[i_dv_decode]
            if isinstance(dv, (OrdinalVariable, CategoricalVariable)):
                values = dv.values
                decoded_values = [values[int(x_ij)] for x_ij in x_encoded]
                return decoded_values

            # No need to decode integer or float variables
            return list(x_encoded)

        # Decode one design variable
        if i_dv is not None:
            if len(x.shape) == 2:
                x_i = x[:, i_dv]
            elif len(x.shape) == 1:
                x_i = x
            else:
                raise ValueError("Expected either 1 or 2-dimensional matrix!")

            # No need to decode for integer or float variable
            return _decode_dv(x_i, i_dv_decode=i_dv)

        # Decode design vectors
        n_dv = self.n_dv
        is_1d = len(x.shape) == 1
        x_mat = np.atleast_2d(x)
        if x_mat.shape[1] != n_dv:
            raise ValueError(
                f"Incorrect number of inputs, expected {n_dv} design variables, received {x_mat.shape[1]}"
            )

        decoded_des_vars = [_decode_dv(x_mat[:, i], i_dv_decode=i) for i in range(n_dv)]
        decoded_des_vectors = [
            [decoded_des_vars[i][ix] for i in range(n_dv)]
            for ix in range(x_mat.shape[0])
        ]
        return decoded_des_vectors[0] if is_1d else decoded_des_vectors

    def sample_valid_x(
        self, n: int, unfolded=False, random_state=None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Sample n design vectors and additionally return the is_acting matrix.

        Parameters
        ----------
        n: int
           - Number of samples to generate
        unfolded: bool
           - Whether to return the samples in unfolded space (each categorical level gets its own dimension)

        Returns
        -------
        x: np.ndarray [n, dim]
           - Valid design vectors
        is_acting: np.ndarray [n, dim]
           - Boolean matrix specifying for each variable whether it is acting or non-acting
        """

        # Sample from the design space

        x, is_acting = self._sample_valid_x(n, random_state=random_state)

        # Check conditionally-acting status
        if np.any(~is_acting[:, ~self.is_conditionally_acting]):
            raise RuntimeError("Unconditionally acting variables cannot be non-acting!")

        # Unfold if needed
        if unfolded:
            x, is_acting = self.unfold_x(x, is_acting)

        return x, is_acting

    def get_x_limits(self) -> list:
        """Returns the variable limit definitions in SMT < 2.0 style"""
        return [dv.get_limits() for dv in self.design_variables]

    def get_num_bounds(self):
        """
        Get bounds for the design space.

        Returns
        -------
        np.ndarray [nx, 2]
           - Bounds of each dimension
        """
        return np.array([(dv.lower, dv.upper) for dv in self.design_variables])

    def get_unfolded_num_bounds(self):
        """
        Get bounds for the unfolded continuous space.

        Returns
        -------
        np.ndarray [nx cont, 2]
           - Bounds of each dimension where limits for categorical variables are expanded to [0, 1]
        """
        unfolded_x_limits = []
        for dv in self.design_variables:
            if isinstance(dv, CategoricalVariable):
                unfolded_x_limits += [[0, 1]] * dv.n_values

            elif isinstance(dv, OrdinalVariable):
                # Note that this interpretation is slightly different from the original mixed_integer implementation in
                # smt: we simply map ordinal values to integers, instead of converting them to integer literals
                # This ensures that each ordinal value gets sampled evenly, also if the values themselves represent
                # unevenly spaced (e.g. log-spaced) values
                unfolded_x_limits.append([dv.lower, dv.upper])

            else:
                unfolded_x_limits.append(dv.get_limits())

        return np.array(unfolded_x_limits).astype(float)

    def fold_x(
        self,
        x: np.ndarray,
        is_acting: np.ndarray = None,
        fold_mask: np.ndarray = None,
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """
        Fold x and optionally is_acting. Folding reverses the one-hot encoding of categorical variables applied by
        unfolding.

        Parameters
        ----------
        x: np.ndarray [n, dim_unfolded]
           - Unfolded samples
        is_acting: np.ndarray [n, dim_unfolded]
           - Boolean matrix specifying for each unfolded variable whether it is acting or non-acting
        fold_mask: np.ndarray [dim_folded]
           - Mask specifying which design variables to apply folding for

        Returns
        -------
        x_folded: np.ndarray [n, dim]
           - Folded samples
        is_acting_folded: np.ndarray [n, dim]
           - (Optional) boolean matrix specifying for each folded variable whether it is acting or non-acting
        """

        # Get number of unfolded dimension
        x = np.atleast_2d(x)
        x_folded = np.zeros((x.shape[0], len(self.design_variables)))
        is_acting_folded = (
            np.ones(x_folded.shape, dtype=bool) if is_acting is not None else None
        )

        i_x_unfold = 0
        for i, dv in enumerate(self.design_variables):
            if (isinstance(dv, CategoricalVariable)) and (
                fold_mask is None or fold_mask[i]
            ):
                n_dim_cat = dv.n_values

                # Categorical values are folded by reversed one-hot encoding:
                # [[1, 0, 0], [0, 1, 0], [0, 0, 1]] --> [0, 1, 2].T
                x_cat_unfolded = x[:, i_x_unfold : i_x_unfold + n_dim_cat]
                value_index = np.argmax(x_cat_unfolded, axis=1)
                x_folded[:, i] = value_index

                # The is_acting matrix is repeated column-wise, so we can just take the first column
                if is_acting is not None:
                    is_acting_folded[:, i] = is_acting[:, i_x_unfold]

                i_x_unfold += n_dim_cat

            else:
                x_folded[:, i] = x[:, i_x_unfold]
                if is_acting is not None:
                    is_acting_folded[:, i] = is_acting[:, i_x_unfold]
                i_x_unfold += 1

        return x_folded, is_acting_folded

    def unfold_x(
        self, x: np.ndarray, is_acting: np.ndarray = None, fold_mask: np.ndarray = None
    ) -> Tuple[np.ndarray, Optional[np.ndarray]]:
        """
        Unfold x and optionally is_acting. Unfolding creates one extra dimension for each categorical variable using
        one-hot encoding.

        Parameters
        ----------
        x: np.ndarray [n, dim]
           - Folded samples
        is_acting: np.ndarray [n, dim]
           - Boolean matrix specifying for each variable whether it is acting or non-acting
        fold_mask: np.ndarray [dim_folded]
           - Mask specifying which design variables to apply folding for

        Returns
        -------
        x_unfolded: np.ndarray [n, dim_unfolded]
           - Unfolded samples
        is_acting_unfolded: np.ndarray [n, dim_unfolded]
           - (Optional) boolean matrix specifying for each unfolded variable whether it is acting or non-acting
        """

        # Get number of unfolded dimension
        n_dim_unfolded = self._get_n_dim_unfolded()
        x = np.atleast_2d(x)
        x_unfolded = np.zeros((x.shape[0], n_dim_unfolded))
        is_acting_unfolded = (
            np.ones(x_unfolded.shape, dtype=bool) if is_acting is not None else None
        )

        i_x_unfold = 0
        for i, dv in enumerate(self.design_variables):
            if isinstance(dv, CategoricalVariable) and (
                fold_mask is None or fold_mask[i]
            ):
                n_dim_cat = dv.n_values
                x_cat = x_unfolded[:, i_x_unfold : i_x_unfold + n_dim_cat]

                # Categorical values are unfolded by one-hot encoding:
                # [0, 1, 2].T --> [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
                x_i_int = x[:, i].astype(int)
                for i_level in range(n_dim_cat):
                    has_value_mask = x_i_int == i_level
                    x_cat[has_value_mask, i_level] = 1

                # The is_acting matrix is simply repeated column-wise
                if is_acting is not None:
                    is_acting_unfolded[:, i_x_unfold : i_x_unfold + n_dim_cat] = (
                        np.tile(is_acting[:, [i]], (1, n_dim_cat))
                    )

                i_x_unfold += n_dim_cat

            else:
                x_unfolded[:, i_x_unfold] = x[:, i]
                if is_acting is not None:
                    is_acting_unfolded[:, i_x_unfold] = is_acting[:, i]
                i_x_unfold += 1

        x_unfolded = x_unfolded[:, :i_x_unfold]
        if is_acting is not None:
            is_acting_unfolded = is_acting_unfolded[:, :i_x_unfold]

        return x_unfolded, is_acting_unfolded

    def _get_n_dim_unfolded(self) -> int:
        return sum(
            [
                dv.n_values if isinstance(dv, CategoricalVariable) else 1
                for dv in self.design_variables
            ]
        )

    @staticmethod
    def _round_equally_distributed(x_cont, lower: int, upper: int):
        """
        To ensure equal distribution of continuous values to discrete values, we first stretch-out the continuous values
        to extend to 0.5 beyond the integer limits and then round. This ensures that the values at the limits get a
        large-enough share of the continuous values.
        """

        x_cont[x_cont < lower] = lower
        x_cont[x_cont > upper] = upper

        diff = upper - lower
        x_stretched = (x_cont - lower) * ((diff + 0.9999) / (diff + 1e-16)) - 0.5
        return np.round(x_stretched) + lower

    """IMPLEMENT FUNCTIONS BELOW"""

    def _get_design_variables(self) -> List[DesignVariable]:
        """Return the design variables defined in this design space if not provided upon initialization of the class"""

    def _is_conditionally_acting(self) -> np.ndarray:
        """
        Return for each design variable whether it is conditionally acting or not. A design variable is conditionally
        acting if it MAY be non-acting.

        Returns
        -------
        is_conditionally_acting: np.ndarray [dim]
            - Boolean vector specifying for each design variable whether it is conditionally acting
        """
        raise NotImplementedError

    def _correct_get_acting(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Correct the given matrix of design vectors and return the corrected vectors and the is_acting matrix.

        Parameters
        ----------
        x: np.ndarray [n_obs, dim]
           - Input variables

        Returns
        -------
        x_corrected: np.ndarray [n_obs, dim]
           - Corrected and imputed input variables
        is_acting: np.ndarray [n_obs, dim]
           - Boolean matrix specifying for each variable whether it is acting or non-acting
        """
        raise NotImplementedError

    def _sample_valid_x(
        self,
        n: int,
        random_state=None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Sample n design vectors and additionally return the is_acting matrix.

        Returns
        -------
        x: np.ndarray [n, dim]
           - Valid design vectors
        is_acting: np.ndarray [n, dim]
           - Boolean matrix specifying for each variable whether it is acting or non-acting
        """
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError


VarValueType = Union[int, str, List[Union[int, str]]]


def raise_config_space():
    raise RuntimeError("Dependencies are not installed, run: pip install smt[cs]")


class DesignSpace(BaseDesignSpace):
    """
    Class for defining a (hierarchical) design space by defining design variables, defining decreed variables
    (optional), and adding value constraints (optional).

    Numerical bounds can be requested using `get_num_bounds()`.
    If needed, it is possible to get the legacy SMT < 2.0 `xlimits` format using `get_x_limits()`.

    Parameters
    ----------
    design_variables: list[DesignVariable]
       - The list of design variables: FloatVariable, IntegerVariable, OrdinalVariable, or CategoricalVariable

    Examples
    --------
    Instantiate the design space with all its design variables:

    >>> from smt.utils.design_space import *
    >>> ds = DesignSpace([
    >>>     CategoricalVariable(['A', 'B']),  # x0 categorical: A or B; order is not relevant
    >>>     OrdinalVariable(['C', 'D', 'E']),  # x1 ordinal: C, D or E; order is relevant
    >>>     IntegerVariable(0, 2),  # x2 integer between 0 and 2 (inclusive): 0, 1, 2
    >>>     FloatVariable(0, 1),  # c3 continuous between 0 and 1
    >>> ])
    >>> assert len(ds.design_variables) == 4

    You can define decreed variables (conditional activation):

    >>> ds.declare_decreed_var(decreed_var=1, meta_var=0, meta_value='A')  # Activate x1 if x0 == A

    Decreed variables can be chained (however no cycles and no "diamonds" are supported):
    Note: only if ConfigSpace is installed! pip install smt[cs]
    >>> ds.declare_decreed_var(decreed_var=2, meta_var=1, meta_value=['C', 'D'])  # Activate x2 if x1 == C or D

    If combinations of values between two variables are not allowed, this can be done using a value constraint:
    Note: only if ConfigSpace is installed! pip install smt[cs]
    >>> ds.add_value_constraint(var1=0, value1='A', var2=2, value2=[0, 1])  # Forbid x0 == A && x2 == 0 or 1

    After defining everything correctly, you can then use the design space object to correct design vectors and get
    information about which design variables are acting:

    >>> x_corr, is_acting = ds.correct_get_acting(np.array([
    >>>     [0, 0, 2, .25],
    >>>     [0, 2, 1, .75],
    >>> ]))
    >>> assert np.all(x_corr == np.array([
    >>>     [0, 0, 2, .25],
    >>>     [0, 2, 0, .75],
    >>> ]))
    >>> assert np.all(is_acting == np.array([
    >>>     [True, True, True, True],
    >>>     [True, True, False, True],  # x2 is not acting if x1 != C or D (0 or 1)
    >>> ]))

    It is also possible to randomly sample design vectors conforming to the constraints:

    >>> x_sampled, is_acting_sampled = ds.sample_valid_x(100)

    You can also instantiate a purely-continuous design space from bounds directly:

    >>> continuous_design_space = DesignSpace([(0, 1), (0, 2), (.5, 5.5)])
    >>> assert continuous_design_space.n_dv == 3

    If needed, it is possible to get the legacy design space definition format:

    >>> xlimits = ds.get_x_limits()
    >>> cont_bounds = ds.get_num_bounds()
    >>> unfolded_cont_bounds = ds.get_unfolded_num_bounds()

    """

    def __init__(
        self,
        design_variables: Union[List[DesignVariable], list, np.ndarray],
        random_state=None,
    ):
        self.sampler = None

        # Assume float variable bounds as inputs
        def _is_num(val):
            try:
                float(val)
                return True
            except ValueError:
                return False

        if len(design_variables) > 0 and not isinstance(
            design_variables[0], DesignVariable
        ):
            converted_dvs = []
            for bounds in design_variables:
                if len(bounds) != 2 or not _is_num(bounds[0]) or not _is_num(bounds[1]):
                    raise RuntimeError(
                        f"Expecting either a list of DesignVariable objects or float variable "
                        f"bounds! Unrecognized: {bounds!r}"
                    )
                converted_dvs.append(FloatVariable(bounds[0], bounds[1]))
            design_variables = converted_dvs

        self.random_state = random_state  # For testing

        self._cs = None
        self._cs_cate = None
        if HAS_CONFIG_SPACE:
            cs_vars = {}
            cs_vars_cate = {}
            self.isinteger = False
            for i, dv in enumerate(design_variables):
                name = f"x{i}"
                if isinstance(dv, FloatVariable):
                    cs_vars[name] = UniformFloatHyperparameter(
                        name, lower=dv.lower, upper=dv.upper
                    )
                    cs_vars_cate[name] = UniformFloatHyperparameter(
                        name, lower=dv.lower, upper=dv.upper
                    )
                elif isinstance(dv, IntegerVariable):
                    cs_vars[name] = FixedIntegerParam(
                        name, lower=dv.lower, upper=dv.upper
                    )
                    listvalues = []
                    for i in range(int(dv.upper - dv.lower + 1)):
                        listvalues.append(str(int(i + dv.lower)))
                    cs_vars_cate[name] = CategoricalHyperparameter(
                        name, choices=listvalues
                    )
                    self.isinteger = True
                elif isinstance(dv, OrdinalVariable):
                    cs_vars[name] = OrdinalHyperparameter(name, sequence=dv.values)
                    cs_vars_cate[name] = CategoricalHyperparameter(
                        name, choices=dv.values
                    )

                elif isinstance(dv, CategoricalVariable):
                    cs_vars[name] = CategoricalHyperparameter(name, choices=dv.values)
                    cs_vars_cate[name] = CategoricalHyperparameter(
                        name, choices=dv.values
                    )

                else:
                    raise ValueError(f"Unknown variable type: {dv!r}")
            seed = self._to_seed(random_state)

            self._cs = NoDefaultConfigurationSpace(space=cs_vars, seed=seed)
            ## Fix to make constraints work correctly with either IntegerVariable or OrdinalVariable
            ## ConfigSpace is malfunctioning
            self._cs_cate = NoDefaultConfigurationSpace(space=cs_vars_cate, seed=seed)

        # dict[int, dict[any, list[int]]]: {meta_var_idx: {value: [decreed_var_idx, ...], ...}, ...}
        self._meta_vars = {}
        self._is_decreed = np.zeros((len(design_variables),), dtype=bool)

        super().__init__(design_variables)

    def declare_decreed_var(
        self, decreed_var: int, meta_var: int, meta_value: VarValueType
    ):
        """
        Define a conditional (decreed) variable to be active when the meta variable has (one of) the provided values.

        Parameters
        ----------
        decreed_var: int
           - Index of the conditional variable (the variable that is conditionally active)
        meta_var: int
           - Index of the meta variable (the variable that determines whether the conditional var is active)
        meta_value: int | str | list[int|str]
           - The value or list of values that the meta variable can have to activate the decreed var
        """

        # ConfigSpace implementation
        if self._cs is not None:
            # Get associated parameters
            decreed_param = self._get_param(decreed_var)
            meta_param = self._get_param(meta_var)

            # Add a condition that checks for equality (if single value given) or in-collection (if sequence given)
            if isinstance(meta_value, Sequence):
                condition = InCondition(decreed_param, meta_param, meta_value)
            else:
                condition = EqualsCondition(decreed_param, meta_param, meta_value)

            ## Fix to make constraints work correctly with either IntegerVariable or OrdinalVariable
            ## ConfigSpace is malfunctioning
            self._cs.add_condition(condition)
            decreed_param = self._get_param2(decreed_var)
            meta_param = self._get_param2(meta_var)
            # Add a condition that checks for equality (if single value given) or in-collection (if sequence given)
            if isinstance(meta_value, Sequence):
                try:
                    condition = InCondition(
                        decreed_param,
                        meta_param,
                        list(np.atleast_1d(np.array(meta_value, dtype=str))),
                    )
                except ValueError:
                    condition = InCondition(
                        decreed_param,
                        meta_param,
                        list(np.atleast_1d(np.array(meta_value, dtype=float))),
                    )
            else:
                try:
                    condition = EqualsCondition(
                        decreed_param, meta_param, str(meta_value)
                    )
                except ValueError:
                    condition = EqualsCondition(decreed_param, meta_param, meta_value)

            self._cs_cate.add_condition(condition)

        # Simplified implementation
        else:
            # Variables cannot be both meta and decreed at the same time
            if self._is_decreed[meta_var]:
                raise RuntimeError(
                    f"Variable cannot be both meta and decreed ({meta_var})!"
                )

            # Variables can only be decreed by one meta var
            if self._is_decreed[decreed_var]:
                raise RuntimeError(f"Variable is already decreed: {decreed_var}")

            # Define meta-decreed relationship
            if meta_var not in self._meta_vars:
                self._meta_vars[meta_var] = {}

            meta_var_obj = self.design_variables[meta_var]
            for value in (
                meta_value if isinstance(meta_value, Sequence) else [meta_value]
            ):
                encoded_value = value
                if isinstance(meta_var_obj, (OrdinalVariable, CategoricalVariable)):
                    if value in meta_var_obj.values:
                        encoded_value = meta_var_obj.values.index(value)

                if encoded_value not in self._meta_vars[meta_var]:
                    self._meta_vars[meta_var][encoded_value] = []
                self._meta_vars[meta_var][encoded_value].append(decreed_var)

        # Mark as decreed (conditionally acting)
        self._is_decreed[decreed_var] = True

    def add_value_constraint(
        self, var1: int, value1: VarValueType, var2: int, value2: VarValueType
    ):
        """
        Define a constraint where two variables cannot have the given values at the same time.

        Parameters
        ----------
        var1: int
           - Index of the first variable
        value1: int | str | list[int|str]
           - Value or values that the first variable is checked against
        var2: int
           - Index of the second variable
        value2: int | str | list[int|str]
           - Value or values that the second variable is checked against
        """
        if self._cs is None:
            raise_config_space()
        # Get parameters
        param1 = self._get_param(var1)
        param2 = self._get_param(var2)
        mixint_types = (UniformIntegerHyperparameter, OrdinalHyperparameter)
        self.has_valcons_ord_int = isinstance(param1, mixint_types) or isinstance(
            param2, mixint_types
        )
        if not (isinstance(param1, UniformFloatHyperparameter)) and not (
            isinstance(param2, UniformFloatHyperparameter)
        ):
            # Add forbidden clauses
            if isinstance(value1, Sequence):
                clause1 = ForbiddenInClause(param1, value1)
            else:
                clause1 = ForbiddenEqualsClause(param1, value1)

            if isinstance(value2, Sequence):
                clause2 = ForbiddenInClause(param2, value2)
            else:
                clause2 = ForbiddenEqualsClause(param2, value2)

            constraint_clause = ForbiddenAndConjunction(clause1, clause2)
            self._cs.add_forbidden_clause(constraint_clause)
        else:
            if value1 in [">", "<"] and value2 in [">", "<"] and value1 != value2:
                if value1 == "<":
                    constraint_clause = ForbiddenLessThanRelation(param1, param2)
                    self._cs.add_forbidden_clause(constraint_clause)
                else:
                    constraint_clause = ForbiddenLessThanRelation(param2, param1)
                    self._cs.add_forbidden_clause(constraint_clause)
            else:
                raise ValueError("Bad definition of DesignSpace.")

        ## Fix to make constraints work correctly with either IntegerVariable or OrdinalVariable
        ## ConfigSpace is malfunctioning
        # Get parameters
        param1 = self._get_param2(var1)
        param2 = self._get_param2(var2)
        # Add forbidden clauses
        if not (isinstance(param1, UniformFloatHyperparameter)) and not (
            isinstance(param2, UniformFloatHyperparameter)
        ):
            if isinstance(value1, Sequence):
                clause1 = ForbiddenInClause(
                    param1, list(np.atleast_1d(np.array(value1, dtype=str)))
                )
            else:
                clause1 = ForbiddenEqualsClause(param1, str(value1))

            if isinstance(value2, Sequence):
                try:
                    clause2 = ForbiddenInClause(
                        param2, list(np.atleast_1d(np.array(value2, dtype=str)))
                    )
                except ValueError:
                    clause2 = ForbiddenInClause(
                        param2, list(np.atleast_1d(np.array(value2, dtype=float)))
                    )
            else:
                try:
                    clause2 = ForbiddenEqualsClause(param2, str(value2))
                except ValueError:
                    clause2 = ForbiddenEqualsClause(param2, value2)

            constraint_clause = ForbiddenAndConjunction(clause1, clause2)
            self._cs_cate.add_forbidden_clause(constraint_clause)

    def _get_param(self, idx):
        try:
            return self._cs.get_hyperparameter(f"x{idx}")
        except KeyError:
            raise KeyError(f"Variable not found: {idx}")

    def _get_param2(self, idx):
        try:
            return self._cs_cate.get_hyperparameter(f"x{idx}")
        except KeyError:
            raise KeyError(f"Variable not found: {idx}")

    @property
    def _cs_var_idx(self):
        """
        ConfigurationSpace applies topological sort when adding conditions, so compared to what we expect the order of
        parameters might have changed.

        This property contains the indices of the params in the ConfigurationSpace.
        """
        names = self._cs.get_hyperparameter_names()
        return np.array(
            [names.index(f"x{ix}") for ix in range(len(self.design_variables))]
        )

    @property
    def _inv_cs_var_idx(self):
        """
        See _cs_var_idx. This function returns the opposite mapping: the positions of our design variables for each
        param.
        """
        return np.array(
            [int(param[1:]) for param in self._cs.get_hyperparameter_names()]
        )

    def _is_conditionally_acting(self) -> np.ndarray:
        # Decreed variables are the conditionally acting variables
        return self._is_decreed

    def _correct_get_acting(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Correct and impute design vectors"""
        x = x.astype(float)
        if self._cs is not None:
            # Normalize value according to what ConfigSpace expects
            self._normalize_x(x)

            # Get corrected Configuration objects by mapping our design vectors
            # to the ordering of the ConfigurationSpace
            inv_cs_var_idx = self._inv_cs_var_idx
            configs = []
            for xi in x:
                configs.append(self._get_correct_config(xi[inv_cs_var_idx]))

            # Convert Configuration objects to design vectors and get the is_active matrix
            return self._configs_to_x(configs)

        # Simplified implementation
        # Correct discrete variables
        x_corr = x.copy()
        self._normalize_x(x_corr, cs_normalize=False)

        # Determine which variables are acting
        is_acting = np.ones(x_corr.shape, dtype=bool)
        is_acting[:, self._is_decreed] = False
        for i, xi in enumerate(x_corr):
            for i_meta, decrees in self._meta_vars.items():
                meta_var_value = xi[i_meta]
                if meta_var_value in decrees:
                    i_decreed_vars = decrees[meta_var_value]
                    is_acting[i, i_decreed_vars] = True

        # Impute non-acting variables
        self._impute_non_acting(x_corr, is_acting)

        return x_corr, is_acting

    def _to_seed(self, random_state=None):
        seed = None
        if isinstance(random_state, int):
            seed = random_state
        elif isinstance(random_state, np.random.RandomState):
            seed = random_state.get_state()[1][0]
        return seed

    def _sample_valid_x(
        self, n: int, random_state=None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Sample design vectors"""
        # Simplified implementation: sample design vectors in unfolded space
        x_limits_unfolded = self.get_unfolded_num_bounds()
        if self.random_state is None:
            self.random_state = random_state

        if self._cs is not None:
            # Sample Configuration objects
            if self.seed is None:
                seed = self._to_seed(random_state)
                self.seed = seed
            self._cs.seed(self.seed)
            if self.seed is not None:
                self.seed += 1
            configs = self._cs.sample_configuration(n)
            if n == 1:
                configs = [configs]
            # Convert Configuration objects to design vectors and get the is_active matrix
            return self._configs_to_x(configs)

        else:
            if self.sampler is None:
                self.sampler = LHS(
                    xlimits=x_limits_unfolded,
                    random_state=random_state,
                    criterion="ese",
                )
            x = self.sampler(n)
            # Fold and cast to discrete
            x, _ = self.fold_x(x)
            self._normalize_x(x, cs_normalize=False)
            # Get acting information and impute
            return self.correct_get_acting(x)

    def _get_correct_config(self, vector: np.ndarray) -> Configuration:
        config = Configuration(self._cs, vector=vector)

        # Unfortunately we cannot directly ask which parameters SHOULD be active
        # https://github.com/automl/ConfigSpace/issues/253#issuecomment-1513216665
        # Therefore, we temporarily fix it with a very dirty workaround: catch the error raised in check_configuration
        # to find out which parameters should be inactive
        while True:
            try:
                ## Fix to make constraints work correctly with either IntegerVariable or OrdinalVariable
                ## ConfigSpace is malfunctioning
                if self.isinteger and self.has_valcons_ord_int:
                    vector2 = np.copy(vector)
                    self._cs_denormalize_x_ordered(np.atleast_2d(vector2))
                    indvec = 0
                    for hp in self._cs_cate:
                        if (
                            (str(self._cs.get_hyperparameter(hp)).split()[2])
                            == "UniformInteger,"
                            and (
                                str(self._cs_cate.get_hyperparameter(hp)).split()[2][:3]
                            )
                            == "Cat"
                            and not (np.isnan(vector2[indvec]))
                        ):
                            vector2[indvec] = int(vector2[indvec]) - int(
                                str(self._cs_cate.get_hyperparameter(hp)).split()[4][
                                    1:-1
                                ]
                            )
                        indvec += 1
                    self._normalize_x_no_integer(np.atleast_2d(vector2))
                    config2 = Configuration(self._cs_cate, vector=vector2)
                    config2.is_valid_configuration()

                config.is_valid_configuration()
                return config

            except ValueError as e:
                error_str = str(e)
                if "Inactive hyperparameter" in error_str:
                    # Deduce which parameter is inactive
                    inactive_param_name = error_str.split("'")[1]
                    param_idx = self._cs.get_idx_by_hyperparameter_name(
                        inactive_param_name
                    )

                    # Modify the vector and create a new Configuration
                    vector = config.get_array().copy()
                    vector[param_idx] = np.nan
                    config = Configuration(self._cs, vector=vector)

                # At this point, the parameter active statuses are set correctly, so we only need to correct the
                # configuration to one that does not violate the forbidden clauses
                elif isinstance(e, ForbiddenValueError):
                    if self.seed is None:
                        seed = self._to_seed(self.random_state)
                        self.seed = seed
                    if not (self.has_valcons_ord_int):
                        return get_random_neighbor(config, seed=self.seed)
                    else:
                        vector = config.get_array().copy()
                        indvec = 0
                        vector2 = np.copy(vector)
                        ## Fix to make constraints work correctly with either IntegerVariable or OrdinalVariable
                        ## ConfigSpace is malfunctioning
                        for hp in self._cs_cate:
                            if (
                                str(self._cs_cate.get_hyperparameter(hp)).split()[2][:3]
                            ) == "Cat" and not (np.isnan(vector2[indvec])):
                                vector2[indvec] = int(vector2[indvec])
                            indvec += 1

                        config2 = Configuration(self._cs_cate, vector=vector2)
                        config3 = get_random_neighbor(config2, seed=self.seed)
                        vector3 = config3.get_array().copy()
                        config4 = Configuration(self._cs, vector=vector3)
                        return config4
                else:
                    raise

    def _configs_to_x(
        self, configs: List["Configuration"]
    ) -> Tuple[np.ndarray, np.ndarray]:
        x = np.zeros((len(configs), len(self.design_variables)))
        is_acting = np.zeros(x.shape, dtype=bool)
        if len(configs) == 0:
            return x, is_acting

        cs_var_idx = self._cs_var_idx
        for i, config in enumerate(configs):
            x[i, :] = config.get_array()[cs_var_idx]

        # De-normalize continuous and integer variables
        self._cs_denormalize_x(x)

        # Set is_active flags and impute x
        is_acting = np.isfinite(x)
        self._impute_non_acting(x, is_acting)

        return x, is_acting

    def _impute_non_acting(self, x: np.ndarray, is_acting: np.ndarray):
        for i, dv in enumerate(self.design_variables):
            if isinstance(dv, FloatVariable):
                # Impute continuous variables to the mid of their bounds
                x[~is_acting[:, i], i] = 0.5 * (dv.upper - dv.lower)

            else:
                # Impute discrete variables to their lower bounds
                lower = 0
                if isinstance(dv, (IntegerVariable, OrdinalVariable)):
                    lower = dv.lower

                x[~is_acting[:, i], i] = lower

    def _normalize_x(self, x: np.ndarray, cs_normalize=True):
        for i, dv in enumerate(self.design_variables):
            if isinstance(dv, FloatVariable):
                if cs_normalize:
                    dv.lower = min(np.min(x[:, i]), dv.lower)
                    dv.upper = max(np.max(x[:, i]), dv.upper)
                    x[:, i] = np.clip(
                        (x[:, i] - dv.lower) / (dv.upper - dv.lower + 1e-16), 0, 1
                    )

            elif isinstance(dv, IntegerVariable):
                x[:, i] = self._round_equally_distributed(x[:, i], dv.lower, dv.upper)

                if cs_normalize:
                    # After rounding, normalize between 0 and 1, where 0 and 1 represent the stretched bounds
                    x[:, i] = (x[:, i] - dv.lower + 0.49999) / (
                        dv.upper - dv.lower + 0.9999
                    )

    def _normalize_x_no_integer(self, x: np.ndarray, cs_normalize=True):
        ordereddesign_variables = [
            self.design_variables[i] for i in self._inv_cs_var_idx
        ]
        for i, dv in enumerate(ordereddesign_variables):
            if isinstance(dv, FloatVariable):
                if cs_normalize:
                    x[:, i] = np.clip(
                        (x[:, i] - dv.lower) / (dv.upper - dv.lower + 1e-16), 0, 1
                    )

            elif isinstance(dv, (OrdinalVariable, CategoricalVariable)):
                # To ensure equal distribution of continuous values to discrete values, we first stretch-out the
                # continuous values to extend to 0.5 beyond the integer limits and then round. This ensures that the
                # values at the limits get a large-enough share of the continuous values
                x[:, i] = self._round_equally_distributed(x[:, i], dv.lower, dv.upper)

    def _cs_denormalize_x(self, x: np.ndarray):
        for i, dv in enumerate(self.design_variables):
            if isinstance(dv, FloatVariable):
                x[:, i] = x[:, i] * (dv.upper - dv.lower) + dv.lower

            elif isinstance(dv, IntegerVariable):
                # Integer values are normalized similarly to what is done in _round_equally_distributed
                x[:, i] = np.round(
                    x[:, i] * (dv.upper - dv.lower + 0.9999) + dv.lower - 0.49999
                )

    def _cs_denormalize_x_ordered(self, x: np.ndarray):
        ordereddesign_variables = [
            self.design_variables[i] for i in self._inv_cs_var_idx
        ]
        for i, dv in enumerate(ordereddesign_variables):
            if isinstance(dv, FloatVariable):
                x[:, i] = x[:, i] * (dv.upper - dv.lower) + dv.lower

            elif isinstance(dv, IntegerVariable):
                # Integer values are normalized similarly to what is done in _round_equally_distributed
                x[:, i] = np.round(
                    x[:, i] * (dv.upper - dv.lower + 0.9999) + dv.lower - 0.49999
                )

    def __str__(self):
        dvs = "\n".join([f"x{i}: {dv!s}" for i, dv in enumerate(self.design_variables)])
        return f"Design space:\n{dvs}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.design_variables!r})"


class NoDefaultConfigurationSpace(ConfigurationSpace):
    """ConfigurationSpace that supports no default configuration"""

    def get_default_configuration(self, *args, **kwargs):
        raise NotImplementedError

    def _check_default_configuration(self, *args, **kwargs):
        pass


class FixedIntegerParam(UniformIntegerHyperparameter):
    def get_neighbors(
        self,
        value: float,
        rs: np.random.RandomState,
        number: int = 4,
        transform: bool = False,
        std: float = 0.2,
    ) -> List[int]:
        # Temporary fix until https://github.com/automl/ConfigSpace/pull/313 is released
        center = self._transform(value)
        lower, upper = self.lower, self.upper
        if upper - lower - 1 < number:
            neighbors = sorted(set(range(lower, upper + 1)) - {center})
            if transform:
                return neighbors
            return self._inverse_transform(np.asarray(neighbors)).tolist()

        return super().get_neighbors(
            value, rs, number=number, transform=transform, std=std
        )



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 30 09:30:40 2023

@author: hvalayer, rlafage

This implementation in SMT is derived from FITC/VarDTC methods from
Sparse GP implementations of GPy project. See https://github.com/SheffieldML/GPy
"""

import numpy as np
from scipy import linalg

from smt.surrogate_models.krg import KRG
from smt.utils.checks import ensure_2d_array
from smt.utils.kriging import differences
from smt.utils.misc import standardization


class SGP(KRG):
    name = "SGP"

    def _initialize(self):
        super()._initialize()

        declare = self.options.declare
        declare(
            "corr",
            "squar_exp",  # gaussian kernel only
            values=("squar_exp"),
            desc="Correlation function type",
            types=(str),
        )
        declare(
            "poly",
            "constant",  # constant mean function
            values=("constant"),
            desc="Regression function type",
            types=(str),
        )
        declare(
            "theta_bounds",
            [1e-6, 1e2],  # upper bound increased compared to kriging-based one
            types=(list, np.ndarray),
            desc="bounds for hyperparameters",
        )
        declare(
            "noise0",
            [1e-2],
            desc="Gaussian noise on observed training data",
            types=(list, np.ndarray),
        )
        declare(
            "hyper_opt",
            "Cobyla",
            values=("Cobyla"),
            desc="Optimiser for hyperparameters optimisation",
            types=str,
        )
        declare(
            "eval_noise",
            True,  # for SGP evaluate noise by default
            types=bool,
            values=(True, False),
            desc="Noise is always evaluated",
        )
        declare(
            "nugget",
            1000.0
            * np.finfo(
                np.double
            ).eps,  # slightly increased compared to kriging-based one
            types=(float),
            desc="a jitter for numerical stability",
        )
        declare(
            "method",
            "FITC",
            values=("FITC", "VFE"),
            desc="Method used by sparse GP model",
            types=(str),
        )
        declare("n_inducing", 10, desc="Number of inducing inputs", types=int)

        supports = self.supports
        supports["derivatives"] = False
        supports["variances"] = True
        supports["variance_derivatives"] = False
        supports["x_hierarchy"] = False

        self.Z = None
        self.woodbury_data = {"vec": None, "inv": None}
        self.optimal_par = {}
        self.optimal_noise = None

    def set_inducing_inputs(self, Z=None, normalize=False):
        """
        Define number of inducing inputs or set the locations manually.
        When Z is not specified then points are picked randomly amongst the inputs training set.

        Parameters
        ----------
        nz : int, optional
            Number of inducing inputs.
        Z : np.ndarray [M,ndim], optional
            Inducing inputs.
        normalize : When Z is given, whether values should be normalized
        """
        if Z is None:
            self.nz = self.options["n_inducing"]
            X = self.training_points[None][0][0]  # [nt,nx]
            random_idx = np.random.permutation(self.nt)[: self.nz]
            self.Z = X[random_idx].copy()  # [nz,nx]
        else:
            Z = ensure_2d_array(Z, "Z")
            if self.nx != Z.shape[1]:
                raise ValueError("DimensionError: Z.shape[1] != X.shape[1]")
            self.Z = Z  # [nz,nx]
            if normalize:
                X = self.training_points[None][0][0]  # [nt,nx]
                y = self.training_points[None][0][1]
                self.normalize = True
                (
                    _,
                    _,
                    X_offset,
                    _,
                    X_scale,
                    _,
                ) = standardization(X, y)
                self.Z = (self.Z - X_offset) / X_scale
            else:
                self.normalize = False
            self.nz = Z.shape[0]

    # overload kriging based implementation
    def _new_train(self):
        if self.Z is None:
            self.set_inducing_inputs()

        # make sure the latent function is scalars
        Y = self.training_points[None][0][1]
        _, output_dim = Y.shape
        if output_dim > 1:
            raise NotImplementedError("SGP does not support vector-valued function")

        # make sure the noise is not hetero
        if self.options["use_het_noise"]:
            raise NotImplementedError("SGP does not support heteroscedastic noise")

        # make sure we are using continuous variables only
        if not self.is_continuous:
            raise NotImplementedError("SGP does not support mixed-integer variables")

        # works only with COBYLA because of no likelihood gradients
        if self.options["hyper_opt"] != "Cobyla":
            raise NotImplementedError("SGP works only with COBYLA internal opimizer")

        return super()._new_train()

    # overload kriging based implementation
    def _reduced_likelihood_function(self, theta):
        X = self.training_points[None][0][0]
        Y = self.training_points[None][0][1]
        Z = self.Z

        if self.options["eval_noise"]:
            sigma2 = theta[-2]
            noise = theta[-1]
            theta = theta[0:-2]
        else:
            sigma2 = theta[-1]
            noise = self.options["noise0"]
            theta = theta[0:-1]

        nugget = self.options["nugget"]

        if self.options["method"] == "VFE":
            likelihood, w_vec, w_inv = self._vfe(X, Y, Z, theta, sigma2, noise, nugget)
        else:
            likelihood, w_vec, w_inv = self._fitc(X, Y, Z, theta, sigma2, noise, nugget)

        self.woodbury_data["vec"] = w_vec
        self.woodbury_data["inv"] = w_inv

        params = {
            "theta": theta,
            "sigma2": sigma2,
        }

        return likelihood, params

    def _reduced_likelihood_gradient(self, theta):
        raise NotImplementedError(
            "SGP gradient of reduced likelihood not implemented yet"
        )

    def _reduced_likelihood_hessian(self, theta):
        raise NotImplementedError(
            "SGP hessian of reduced likelihood not implemented yet"
        )

    def _compute_K(self, A: np.ndarray, B: np.ndarray, theta, sigma2):
        """
        Compute the covariance matrix K between A and B
        """
        # Compute pairwise componentwise L1-distances between A and B
        dx = differences(A, B)
        d = self._componentwise_distance(dx)
        # Compute the correlation vector r and matrix R
        r = self._correlation_types[self.options["corr"]](theta, d)
        R = r.reshape(A.shape[0], B.shape[0])
        # Compute the covariance matrix K
        K = sigma2 * R
        return K

    def _fitc(self, X, Y, Z, theta, sigma2, noise, nugget):
        """
        FITC method implementation.
        """

        # Compute: diag(Knn), Kmm and Kmn
        Knn = np.full(self.nt, sigma2)
        Kmm = self._compute_K(Z, Z, theta, sigma2) + np.eye(self.nz) * nugget
        Kmn = self._compute_K(Z, X, theta, sigma2)

        # Compute (lower) Cholesky decomposition: Kmm = U U^T
        U = linalg.cholesky(Kmm, lower=True)

        # Compute (upper) Cholesky decomposition: Qnn = V^T V
        Ui = linalg.inv(U)
        V = Ui @ Kmn

        # Assumption on the gaussian noise on training outputs
        eta2 = noise

        # Compute diagonal correction: nu = Knn_diag - Qnn_diag + \eta^2
        nu = Knn - np.sum(np.square(V), 0) + eta2
        # Compute beta, the effective noise precision
        beta = 1.0 / nu

        # Compute (lower) Cholesky decomposition: A = I + V diag(beta) V^T = L L^T
        A = np.eye(self.nz) + V * beta @ V.T
        L = linalg.cholesky(A, lower=True)
        Li = linalg.inv(L)

        # Compute a and b
        a = np.einsum("ij,i->ij", Y, beta)  # avoid reshape for mat-vec multiplication
        b = Li @ V @ a

        # Compute marginal log-likelihood
        likelihood = -0.5 * (
            # num_data * np.log(2.0 * np.pi)   # constant term ignored in reduced likelihood
            +np.sum(np.log(nu))
            + 2.0 * np.sum(np.log(np.diag(L)))
            + a.T @ Y
            - np.einsum("ij,ij->", b, b)
        )

        # Store Woodbury vectors for prediction step
        LiUi = Li @ Ui
        LiUiT = LiUi.T
        woodbury_vec = LiUiT @ b
        woodbury_inv = Ui.T @ Ui - LiUiT @ LiUi

        return likelihood, woodbury_vec, woodbury_inv

    def _vfe(self, X, Y, Z, theta, sigma2, noise, nugget):
        """
        VFE method implementation.
        """

        # Assume zero mean function
        mean = 0
        Y -= mean

        # Compute: diag(Knn), Kmm and Kmn
        Kmm = self._compute_K(Z, Z, theta, sigma2) + np.eye(self.nz) * nugget
        Kmn = self._compute_K(Z, X, theta, sigma2)

        # Compute (lower) Cholesky decomposition: Kmm = U U^T
        U = linalg.cholesky(Kmm, lower=True)

        # Compute (upper) Cholesky decomposition: Qnn = V^T V
        Ui = linalg.inv(U)
        V = Ui @ Kmn

        # Compute beta, the effective noise precision
        beta = 1.0 / np.fmax(noise, nugget)

        # Compute A = beta * V @ V.T
        A = beta * V @ V.T

        # Compute (lower) Cholesky decomposition: B = I + A = L L^T
        B = np.eye(self.nz) + A
        L = linalg.cholesky(B, lower=True)
        Li = linalg.inv(L)

        # Compute b
        b = beta * Li @ V @ Y

        # Compute log-marginal likelihood
        likelihood = -0.5 * (
            # self.nt * np.log(2.0 * np.pi)   # constant term ignored in reduced likelihood
            -self.nt * np.log(beta)
            + 2.0 * np.sum(np.log(np.diag(L)))
            + beta * np.einsum("ij,ij->", Y, Y)
            - b.T @ b
            + self.nt * beta * sigma2
            - np.trace(A)
        )

        # Store Woodbury vectors for prediction step
        LiUi = Li @ Ui
        Bi = np.eye(self.nz) + Li.T @ Li
        woodbury_vec = LiUi.T @ b
        woodbury_inv = Ui.T @ Bi @ Ui

        return likelihood, woodbury_vec, woodbury_inv

    # overload kriging based implementation
    def _predict_values(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        """
        Evaluates the model at a set of points using the Woodbury vector.
        """
        Kx = self._compute_K(x, self.Z, self.optimal_theta, self.optimal_sigma2)
        mu = Kx @ self.woodbury_data["vec"]
        return mu

    # overload kriging based implementation
    def _predict_variances(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        """
        Evaluates the model at a set of points using the inverse Woodbury vector.
        """
        Kx = self._compute_K(self.Z, x, self.optimal_theta, self.optimal_sigma2)
        Kxx = np.full(x.shape[0], self.optimal_sigma2)
        var = (Kxx - np.sum(np.dot(self.woodbury_data["inv"].T, Kx) * Kx, 0))[:, None]
        var = np.clip(var, 1e-15, np.inf)
        var += self.optimal_noise
        return var



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.
"""

import numpy as np

VALID_LINE_SEARCHES = ("backtracking", "bracketed", "quadratic", "cubic", "null")


def get_line_search_class(line_search):
    if line_search == "backtracking":
        return BacktrackingLineSearch
    elif line_search == "bracketed":
        return BracketedLineSearch
    elif line_search == "quadratic":
        return QuadraticLineSearch
    elif line_search == "cubic":
        return CubicLineSearch
    elif line_search == "null":
        return NullLineSearch


class LineSearch(object):
    """
    Base line search class.
    """

    def __init__(self, x, dx, func, grad, u1=1.0e-4, u2=0.9):
        """
        Initialize all attributes for the given problem.

        Arguments
        ---------
        x : ndarray[:]
            Vector representing the current location in the n-D space.
        dx : ndarray[:]
            Search direction.
        func : function
            scalar function of x.
        grad : function
            vector function that yields the gradient of func.
        u1 : float
            Parameter in the sufficient decrease criterion to ensure non-zero decrease.
        u2 : float
            Parameter in the curvature criterion to ensure gradient norm decreases.
        """
        self.x = x
        self.dx = dx
        self.func = func
        self.grad = grad

        self.u1 = u1
        self.u2 = u2

        self.phi_0 = self._phi(0.0)
        self.dphi_0 = self._dphi(0.0)

    def _phi(self, a):
        """
        Function in terms of alpha (a).

        phi(a) = func(x + a dx)
        """
        return self.func(self.x + a * self.dx)

    def _dphi(self, a):
        """
        Derivative of phi w.r.t. alpha (a).
        """
        return np.dot(self.grad(self.x + a * self.dx), self.dx)

    def _func_decreased(self, a):
        """
        Check sufficient decrease criterion.
        """
        return self._phi(a) <= self.phi_0 + self.u1 * a * self.dphi_0

    def _grad_decreased(self, a):
        """
        Check curvature criterion.
        """
        return np.abs(self._dphi(a)) <= np.abs(self.u2 * self.dphi_0)


class NullLineSearch(object):
    """
    Base line search class.
    """

    def __init__(self, x, dx, func, grad, u1=1.0e-4, u2=0.9):
        """
        Initialize all attributes for the given problem.

        Arguments
        ---------
        x : ndarray[:]
            Vector representing the current location in the n-D space.
        dx : ndarray[:]
            Search direction.
        func : function
            scalar function of x.
        grad : function
            vector function that yields the gradient of func.
        u1 : float
            Parameter in the sufficient decrease criterion to ensure non-zero decrease.
        u2 : float
            Parameter in the curvature criterion to ensure gradient norm decreases.
        """
        self.x = x
        self.dx = dx

    def __call__(self, initial_a=1):
        return self.x + initial_a * self.dx


class BacktrackingLineSearch(LineSearch):
    """
    Simple backtracking line search enforcing only sufficient decrease.
    """

    def __call__(self, initial_a=1.0, rho=0.5):
        a = initial_a
        while not self._func_decreased(a):
            a *= rho
        return self.x + a * self.dx


class BracketedLineSearch(LineSearch):
    """
    Base class for line search algorithms enforcing the Strong Wolfe conditions.
    """

    def __call__(self, initial_a=1):
        a1 = 0
        a2 = initial_a
        p1 = self._phi(a1)
        p2 = self._phi(a2)
        dp1 = self._dphi(a1)
        dp2 = self._dphi(a2)

        for ind in range(20):
            if not self._func_decreased(a2) or p2 > p1:
                # We've successfully bracketed if
                # 1. The function value is greater than at a=0
                # 2. The function value has increased from the previous iteration
                return self._zoom(a1, p1, dp1, a2, p2, dp2)

            if self._grad_decreased(a2):
                # At this point, the func decrease condition is satisfied,
                # so if the grad decrease also is satisfied, we're done.
                return self.x + a2 * self.dx
            elif dp2 >= 0:
                # If only the func decrease is satisfied, but the phi' is positive
                # we've successfully bracketed.
                return self._zoom(a2, p2, dp2, a1, p1, dp1)
            else:
                # Otherwise, we're lower than initial f and previous f,
                # and the slope is still negative and steeper than initial.
                # We can get more aggressive and increase the step.
                a1 = a2
                p1 = p2
                dp1 = dp2
                a2 = a2 * 1.5
                p2 = self._phi(a2)
                dp2 = self._dphi(a2)

    def _zoom(self, a1, p1, dp1, a2, p2, dp2):
        """
        Find a solution in the interval, [a1, a2], assuming that phi(a1) < phi(a2).
        """
        while True:
            a, p, dp = self._compute_minimum(a1, p1, dp1, a2, p2, dp2)

            if not self._func_decreased(a) or p > p1:
                # If still lower than initial f or still higher than low
                # then make this the new high.
                a2 = a
                p2 = p
                dp2 = dp
            else:
                if self._grad_decreased(a):
                    # Both conditions satisfied, so we're done.
                    return self.x + a * self.dx
                elif dp * (a2 - a1) >= 0:
                    # We have a new low and the slope has the right sign.
                    a2 = a1
                    p2 = p1
                    dp2 = dp1
                a1 = a
                p1 = p
                dp1 = dp

    def _compute_minimum(self, a1, p1, dp1, a2, p2, dp2):
        """
        Estimate the minimum as the midpoint.
        """
        a = 0.5 * a1 + 0.5 * a2
        p = self._phi(a)
        dp = self._dphi(a)
        return a, p, dp


class QuadraticLineSearch(BracketedLineSearch):
    """
    Use quadratic interpolation in the zoom method.
    """

    def _compute_minimum(self, a1, p1, dp1, a2, p2, dp2):
        quadratic_mtx = np.zeros((3, 3))
        quadratic_mtx[0, :] = [1.0, a1, a1**2]
        quadratic_mtx[1, :] = [1.0, a2, a2**2]
        quadratic_mtx[2, :] = [0.0, 1.0, 2 * a1]
        c0, c1, c2 = np.linalg.solve(quadratic_mtx, [p1, p2, dp1])

        d0 = c1
        d1 = 2 * c2

        a = -d0 / d1
        p = self._phi(a)
        dp = self._dphi(a)
        return a, p, dp


class CubicLineSearch(BracketedLineSearch):
    """
    Use cubic interpolation in the zoom method.
    """

    def _compute_minimum(self, a1, p1, dp1, a2, p2, dp2):
        cubic_mtx = np.zeros((4, 4))
        cubic_mtx[0, :] = [1.0, a1, a1**2, a1**3]
        cubic_mtx[1, :] = [1.0, a2, a2**2, a2**3]
        cubic_mtx[2, :] = [0.0, 1.0, 2 * a1, 3 * a1**2]
        cubic_mtx[3, :] = [0.0, 1.0, 2 * a2, 3 * a2**2]
        c0, c1, c2, c3 = np.linalg.solve(cubic_mtx, [p1, p2, dp1, dp2])

        d0 = c1
        d1 = 2 * c2
        d2 = 3 * c3
        r1, r2 = np.roots([d2, d1, d0])

        a = None
        p = max(p1, p2)
        if (a1 <= r1 <= a2 or a2 <= r1 <= a1) and np.isreal(r1):
            px = self._phi(r1)
            if px < p:
                a = r1
                p = px
                dp = self._dphi(r1)
        if (a1 <= r2 <= a2 or a2 <= r2 <= a1) and np.isreal(r2):
            px = self._phi(r2)
            if px < p:
                a = r2
                p = px
                dp = self._dphi(r2)

        return a, p, dp



"""
Author: Hugo Reimeringer <hugo.reimeringer@onera.fr>
"""

from sklearn.decomposition import PCA
import numpy as np

from smt.applications.application import SurrogateBasedApplication
from smt.surrogate_models import KRG, KPLS, KPLSK
from smt.utils.checks import ensure_2d_array

PODI_available_models = {
    "KRG": KRG,
    "KPLS": KPLS,
    "KPLSK": KPLSK,
}


class PODI(SurrogateBasedApplication):
    """
    Class for Proper Orthogonal Decomposition and Interpolation (PODI) surrogate models based.

    Attributes
    ----------
    n_modes : int
        Number of kept modes during the POD.
    singular_vectors : np.ndarray
        Singular vectors of the POD.
    singular_values : np.ndarray
        Singular values of the POD.
    interp_coeff : list[SurrogateModel]
        List containing the surrogate models used.

    Examples
    --------
    >>> from smt.applications import PODI
    >>> sm = PODI()
    """

    name = "PODI"

    def _initialize(self) -> None:
        super(PODI, self)._initialize()

        self.available_models_name = []
        for key in PODI_available_models.keys():
            self.available_models_name.append(key)

        self.n_modes = None
        self.singular_vectors = None
        self.singular_values = None

        self.pod_computed = False
        self.interp_options_set = False
        self.training_values_set = False
        self.train_done = False

        self.interp_coeff = None

    @staticmethod
    def choice_n_modes_tol(EV_list: np.ndarray, tol: float) -> int:
        """
        Static method calculating the required number of kept modes to explain at least the intended ratio of variance.

        Parameters
        ----------
        EV_list : np.ndarray
            List of the explained variance of each mode in the POD.
        tol : float
            Desired tolerance for the POD. It is expected to be in ]0, 1].

        Returns
        -------
        n_modes : int
            Kept modes according to the tolerance.
        """

        sum_tot = sum(EV_list)
        sum_ = 0
        for i, ev in enumerate(EV_list):
            sum_ += ev
            EV_ratio = sum_ / sum_tot
            if EV_ratio >= tol:
                return i + 1
        return len(EV_list)

    def compute_pod(
        self,
        database: np.ndarray,
        tol: float = None,
        n_modes: int = None,
        seed: int = None,
    ) -> None:
        """
        Performs the POD.

        Parameters
        ----------
        database : np.ndarray[ny, nt]
            Snapshot matrix. Each column corresponds to a snapshot.
        tol : float
            Desired tolerance for the pod (if n_modes not set).
        n_modes : int
            Desired number of kept modes for the pod (if tol not set).
        seed : int
            seed number which controls random draws for internal optim. (optional)

        Examples
        ----------
        >>> sm.compute_pod(database, tol = 0.99)
        """
        choice_svd = None

        svd = PCA(svd_solver="randomized", random_state=seed)

        if n_modes is not None:
            self.n_modes = n_modes
            choice_svd = "mode"

        if tol is not None:
            if choice_svd is not None:
                raise ValueError(
                    "pod can't use both arguments 'n_modes' and 'tol' at the same time"
                )
            else:
                choice_svd = "tol"

        if choice_svd is None:
            raise ValueError(
                "either one of the arguments 'n_modes' and 'tol' must be specified"
            )

        database = ensure_2d_array(database, "database")

        self.n_snapshot = database.shape[1]
        self.ny = database.shape[0]

        svd.fit(database.T)
        self.singular_vectors = svd.components_.T
        self.singular_values = svd.singular_values_
        EV_list = svd.explained_variance_

        if choice_svd == "tol":
            self.n_modes = PODI.choice_n_modes_tol(EV_list, tol)
        else:
            if self.n_modes > self.n_snapshot:
                raise ValueError(
                    "the number of kept modes can't be superior to the number of data values (snapshots)"
                )
        self.EV_ratio = sum(EV_list[: self.n_modes]) / sum(EV_list)

        self.mean = np.atleast_2d(database.mean(axis=1)).T
        self.basis = self.singular_vectors[:, : self.n_modes]
        self.coeff = np.dot(database.T - self.mean.T, self.basis)

        self.pod_computed = True
        self.interp_options_set = False
        self.training_values_set = False

    def get_singular_vectors(self) -> np.ndarray:
        """
        Getter for the singular vectors of the POD.
        It represents the directions of maximum variance in the data.

        Returns
        -------
        singular_vectors : np.ndarray
            singular vectors of the POD.
        """
        return self.singular_vectors

    def get_singular_values(self) -> np.ndarray:
        """
        Getter for the singular values from the Sigma matrix of the POD.

        Returns
        -------
        singular_values : np.ndarray
            Singular values of the POD.
        """
        return self.singular_values

    def get_ev_ratio(self) -> float:
        """
        Getter for the explained variance ratio with the kept modes.

        Returns
        -------
        EV_ratio : float
            Explained variance ratio with the current kept modes.
        """
        return self.EV_ratio

    def get_n_modes(self) -> int:
        """
        Getter for the number of modes kept during the POD.

        Returns
        -------
        n_modes : int
            number of modes kept during the POD.
        """
        return self.n_modes

    def set_interp_options(
        self, interp_type: str = "KRG", interp_options: list = [{}]
    ) -> None:
        """
        Set the options for the interpolation surrogate models used.
        Only required if a model different than KRG is used or if non-default options are desired for the models.

        Parameters
        ----------
        interp_type : str
            Name of the type of surrogate model that will be used for the whole set.
            By default, the Kriging model is used (KRG).

        interp_options : list[dict]
            List containing dictionnaries for the options.
            The k-th dictionnary corresponds to the options of the k-th interpolation model.
            If the options are common to all surogate models, only a single dictionnary is required in the list.
            The available options can be found in the documentation of the corresponding surrogate models.
            By default, the print_global options are set to 'False'.

        Examples
        --------
        >>> interp_type = "KRG"
        >>> dict1 = {'corr' : 'matern52', 'theta0' : [1e-2]}
        >>> dict2 = {'poly' : 'quadratic'}
        >>> interp_options = [dict1, dict2]
        >>> sm.set_interp_options(interp_type, interp_options)
        """

        if not self.pod_computed:
            raise RuntimeError(
                "'compute_pod' method must have been succesfully executed before trying to set the models options."
            )

        if interp_type not in PODI_available_models.keys():
            raise ValueError(
                f"the surrogate model type should be one of the following : {', '.join(self.available_models_name)}"
            )

        if len(interp_options) == self.n_modes:
            mode_options = "local"
        elif len(interp_options) == 1:
            mode_options = "global"
        else:
            raise ValueError(
                f"expected interp_options of size {self.n_modes} or 1, but got {len(interp_options)}."
            )

        self.interp_coeff = []
        for i in range(self.n_modes):
            if mode_options == "local":
                index = i
            elif mode_options == "global":
                index = 0

            sm_i = PODI_available_models[interp_type](print_global=False)

            for key in interp_options[index].keys():
                sm_i.options[key] = interp_options[index][key]

            self.interp_coeff.append(sm_i)

        self.interp_options_set = True
        self.training_values_set = False

    def set_training_values(self, xt: np.ndarray) -> None:
        """
        Set training data (values).
        If the models' options are still not set, default values are used for the initialization.

        Parameters
        ----------
        xt : np.ndarray[nt, nx]
            The input values for the nt training points.
        """

        xt = ensure_2d_array(xt, "xt")

        if not self.pod_computed:
            raise RuntimeError(
                "'compute_pod' method must have been succesfully executed before trying to set the training values."
            )
        if not self.interp_options_set:
            self.interp_coeff = []
            for i in range(self.n_modes):
                sm_i = PODI_available_models["KRG"](print_global=False)

                self.interp_coeff.append(sm_i)
            self.interp_options_set = True

        self.nt = xt.shape[0]
        self.nx = xt.shape[1]

        if self.nt != self.n_snapshot:
            raise ValueError(
                f"there must be the same amount of train values than data values, {self.nt} != {self.n_snapshot}."
            )

        for i in range(self.n_modes):
            self.interp_coeff[i].set_training_values(xt, self.coeff[:, i])

        self.training_values_set = True

    def train(self) -> None:
        """
        Performs the training of the model.
        """
        if not self.pod_computed:
            raise RuntimeError(
                "'compute_pod' method must have been succesfully executed before trying to train the models."
            )

        if not self.training_values_set:
            raise RuntimeError(
                "the training values should have been set before trying to train the models."
            )

        for interp_coeff in self.interp_coeff:
            interp_coeff.train()

        self.train_done = True

    def get_interp_coeff(self) -> np.ndarray:
        """
        Getter for the list of the interpolation surrogate models used

        Returns
        -------
        interp_coeff : np.ndarray[n_modes]
            List of the kriging models used for the POD coefficients.
        """
        return self.interp_coeff

    def predict_values(self, xn) -> np.ndarray:
        """
        Predict the output values at a set of points.

        Parameters
        ----------
        xn : np.ndarray
            Input values for the prediction points.

        Returns
        -------
        yn : np.ndarray
            Output values at the prediction points.
        """
        if not self.train_done:
            raise RuntimeError(
                "the model should have been trained before trying to make a prediction."
            )

        self.dim_new = xn.shape[1]

        if self.dim_new != self.nx:
            raise ValueError(
                f"the data values and the new values must be the same size, here {self.dim_new} != {self.nx}"
            )

        self.n_new = xn.shape[0]
        mean_coeff_interp = np.zeros((self.n_modes, self.n_new))

        for i in range(self.n_modes):
            mu_i = self.interp_coeff[i].predict_values(xn)
            mean_coeff_interp[i] = mu_i[:, 0]

        y = self.mean + np.dot(self.basis, mean_coeff_interp)

        return y

    def predict_variances(self, xn) -> np.ndarray:
        """
        Predict the variances at a set of points.

        Parameters
        ----------
        xn : np.ndarray[nt, nx]
            Input values for the prediction points.

        Returns
        -------
        s2 : np.ndarray[nt, ny]
            Variances.
        """

        if not self.train_done:
            raise RuntimeError(
                "the model should have been trained before trying to make a prediction"
            )

        dim_new = xn.shape[1]

        if dim_new != self.nx:
            raise ValueError(
                f"the data values and the new values must be the same size, here {self.dim_new} != {self.nx}"
            )

        self.n_new = xn.shape[0]
        var_coeff_interp = np.zeros((self.n_modes, self.n_new))

        for i in range(self.n_modes):
            sigma_i_square = self.interp_coeff[i].predict_variances(xn)
            var_coeff_interp[i] = sigma_i_square[:, 0]

        s2 = np.dot((self.basis**2), var_coeff_interp)

        return s2

    def predict_derivatives(self, xn, kx) -> np.ndarray:
        """
        Predict the dy_dx derivatives at a set of points.

        Parameters
        ----------
        xn : np.ndarray[nt, nx]
            Input values for the prediction points.
        kx : int
            The 0-based index of the input variable with respect to which derivative is desired.

        Returns
        -------
        dy_dx : np.ndarray[nt, ny]
            Derivatives.
        """
        d = kx

        if not self.train_done:
            raise RuntimeError(
                "the model should have been trained before trying to make a prediction"
            )

        dim_new = xn.shape[1]

        if d >= dim_new:
            raise ValueError(
                "the desired derivative kx should correspond to a dimension of the data, here kx is out of bounds."
            )

        if dim_new != self.nx:
            raise ValueError(
                f"the data values and the new values must be the same size, here {self.dim_new} != {self.nx}"
            )

        self.n_new = xn.shape[0]
        deriv_coeff_interp = np.zeros((self.n_modes, self.n_new))

        for i in range(self.n_modes):
            deriv_coeff_interp[i] = self.interp_coeff[i].predict_derivatives(xn, d)[
                :, 0
            ]

        dy_dx = np.dot(self.basis, deriv_coeff_interp)

        return dy_dx



from libcpp.vector cimport vector
from cpython cimport array
import numpy as np
cimport numpy as np


cdef extern from "rbf.hpp":
  cdef cppclass RBF:
    RBF() except +
    void setup(int nx, int nt, int num_dof, int poly_degree, double * d0, double * xt)
    void compute_jac(int n, double* x, double* jac)
    void compute_jac_derivs(int n, int kx, double* x, double* jac)


cdef class PyRBF:

    cdef RBF *thisptr
    def __cinit__(self):
        self.thisptr = new RBF()
    def __dealloc__(self):
        del self.thisptr
    def setup(self,
            int nx, int nt, int num_dof, int poly_degree,
            np.ndarray[double] d0, np.ndarray[double] xt):
        self.thisptr.setup(nx, nt, num_dof, poly_degree, &d0[0], &xt[0])
    def compute_jac(self, int n, np.ndarray[double] x, np.ndarray[double] jac):
        self.thisptr.compute_jac(n, &x[0], &jac[0])
    def compute_jac_derivs(self, int n, int kx, np.ndarray[double] x, np.ndarray[double] jac):
        self.thisptr.compute_jac_derivs(n, kx, &x[0], &jac[0])



"""
Author: Antoine Averland <antoine.averland@onera.fr> and Rémi Lafage <remi.lafage@onera.fr>

This package is distributed under New BSD license.

pyDOE3 sampling methods
"""

import numpy as np
from pyDOE3 import doe_box_behnken, doe_factorial, doe_gsd, doe_plackett_burman

from smt.sampling_methods.sampling_method import SamplingMethod


class PyDoeSamplingMethod(SamplingMethod):
    """
    Base class adapting pyDOE3 designs to SMT SamplingMethod interface
    See https://pydoe3.readthedocs.io/
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.nx = self.options["xlimits"].shape[0]
        self.levels = None

    def _compute(self, nt: int = None):
        """
        Use pydoe3 design to produce [nsamples, nx] matrix
        where nsamples depends on the pyDOE3 method and nx is the dimension of x.
        Warning: In pyDOE3 design setting user requested number of points nt is not used
        """
        xlimits = self.options["xlimits"]
        levels = self.levels

        # Retrieve indices from pyDOE3 design
        doe = np.array(self._compute_doe(), dtype=int)

        # Compute scaled values for each x components
        values = np.zeros((self.nx, max(levels)))
        for i in range(self.nx):
            values[i, 0 : levels[i]] = np.linspace(
                xlimits[i, 0],
                xlimits[i, 1],
                num=levels[i],
            )

        # Use indices to shape the result array and fill it with values
        res = np.zeros(doe.shape)
        i = 0
        for idx in doe:
            for j in range(self.nx):
                res[i, j] = values[j, idx[j]]
            i = i + 1

        return res

    def _compute_doe():
        """Returns a matrix (nsamples, nx) of indices.

        Each indices takes a value in [0, nlevels_i-1] where nlevels_i is
        the number of levels of the ith component of x.
        This method has to be overriden by subclasses"""
        raise NotImplementedError(
            "You have to implement DOE computation method _compute_doe()"
        )


class BoxBehnken(PyDoeSamplingMethod):
    """See https://pydoe3.readthedocs.io/en/latest/rsm.html#box-behnken"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Box Behnken design has 3 levels [-1, 0, 1]
        self.levels = [3] * self.nx  # for

    def _compute_doe(self):
        # Increment Box Behnken levels to get indices [0, 1, 2]
        return doe_box_behnken.bbdesign(self.nx) + 1


class Gsd(PyDoeSamplingMethod):
    """See https://pydoe3.readthedocs.io/en/latest/rsm.html#gsd"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.levels = self.options["levels"]

    def _initialize(self, **kwargs):
        self.options.declare(
            "levels",
            types=list,
            desc="number of factor levels per factor in design",
        )
        self.options.declare(
            "reduction",
            types=int,
            default=2,
            desc="Reduction factor (bigger than 1). Larger `reduction` means fewer experiments \
                  in the design and more possible complementary designs",
        )

    def _compute_doe(self):
        levels = self.options["levels"]
        reduction = self.options["reduction"]

        return doe_gsd.gsd(levels, reduction)


class Factorial(PyDoeSamplingMethod):
    """See https://pydoe3.readthedocs.io/en/latest/factorial.html#general-full-factorial"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.levels = self.options["levels"]

    def _initialize(self, **kwargs):
        self.options.declare(
            "levels",
            types=list,
            desc="number of factor levels per factor in design",
        )

    def _compute_doe(self):
        levels = self.options["levels"]
        return doe_factorial.fullfact(levels)


class PlackettBurman(PyDoeSamplingMethod):
    """See https://pydoe3.readthedocs.io/en/latest/factorial.html#plackett-burman"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Plackett Burman design has 2 levels [-1, 1]
        self.levels = [2] * self.nx

    def _compute_doe(self):
        doe = doe_plackett_burman.pbdesign(self.nx)
        # Change -1 level to get indices [0, 1]
        doe[doe < 0] = 0

        return doe



import numpy as np

from smt.surrogate_models.surrogate_model import SurrogateModel

try:
    import egobox as egx

    GPX_AVAILABLE = True

    REGRESSIONS = {
        "constant": egx.RegressionSpec.CONSTANT,
        "linear": egx.RegressionSpec.LINEAR,
        "quadratic": egx.RegressionSpec.QUADRATIC,
    }
    CORRELATIONS = {
        "abs_exp": egx.CorrelationSpec.ABSOLUTE_EXPONENTIAL,
        "squar_exp": egx.CorrelationSpec.SQUARED_EXPONENTIAL,
        "matern32": egx.CorrelationSpec.MATERN32,
        "matern52": egx.CorrelationSpec.MATERN52,
    }
except ImportError:
    GPX_AVAILABLE = False


class GPX(SurrogateModel):
    name = "GPX"

    def _initialize(self):
        super(GPX, self)._initialize()

        if not GPX_AVAILABLE:
            raise RuntimeError(
                'GPX not available. Please install GPX dependencies with: pip install smt["gpx"]'
            )

        declare = self.options.declare

        declare(
            "poly",
            "constant",
            values=("constant", "linear", "quadratic"),
            desc="Regression function type",
            types=(str),
        )
        declare(
            "corr",
            "squar_exp",
            values=(
                "abs_exp",
                "squar_exp",
                "matern32",
                "matern52",
            ),
            desc="Correlation function type",
        )
        declare(
            "theta0", [1e-2], types=(list, np.ndarray), desc="Initial hyperparameters"
        )
        declare(
            "theta_bounds",
            [1e-6, 2e1],
            types=(list, np.ndarray),
            desc="Bounds for hyperparameters",
        )
        declare(
            "n_start",
            10,
            types=int,
            desc="Number of optimizer runs (multistart method)",
        )
        declare(
            "kpls_dim",
            None,
            types=(type(None), int),
            desc="Number of PLS components used for dimension reduction",
        )
        declare(
            "seed",
            default=42,
            types=int,
            desc="Seed number which controls random draws \
                for internal optim (set by default to get reproductibility)",
        )

        supports = self.supports
        supports["derivatives"] = True
        supports["variances"] = True
        supports["variance_derivatives"] = True

        self._gpx = None

    def _train(self):
        xt, yt = self.training_points[None][0]

        config = {
            "regr_spec": REGRESSIONS[self.options["poly"]],
            "corr_spec": CORRELATIONS[self.options["corr"]],
            "theta_init": np.array(self.options["theta0"]),
            "theta_bounds": np.array([self.options["theta_bounds"]]),
            "n_start": self.options["n_start"],
            "seed": self.options["seed"],
        }
        kpls_dim = self.options["kpls_dim"]
        if kpls_dim:
            config["kpls_dim"] = kpls_dim

        self._gpx = egx.Gpx.builder(**config).fit(xt, yt)

    def _predict_values(self, x):
        return self._gpx.predict(x)

    def _predict_variances(self, x):
        return self._gpx.predict_var(x)

    def _predict_derivatives(self, x, kx):
        return self._gpx.predict_gradients(x)[:, kx : kx + 1]

    def _predict_variance_derivatives(self, x, kx):
        return self._gpx.predict_var_gradients(x)[:, kx : kx + 1]

    def predict_gradients(self, x):
        """Predict derivatives wrt to all x components (eg the gradient)
        at n points given as [n, nx] matrix where nx is the dimension of x.
        Returns all gradients at the given x points as a [n, nx] matrix
        """
        return self._gpx.predict_gradients(x)

    def predict_variance_gradients(self, x):
        """Predict variance derivatives wrt to all x components (eg the variance gradient)
        at n points given as [n, nx] matrix where nx is the dimension of x.
        Returns all variance gradients at the given x points as a [n, nx] matrix"""
        return self._gpx.predict_var_gradients(x)



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

N-dimensional Rosenbrock problem.
"""

import numpy as np

from smt.problems.problem import Problem
from smt.problems.reduced_problem import ReducedProblem
from smt.problems.rosenbrock import Rosenbrock
from smt.utils.options_dictionary import OptionsDictionary


class NdimRosenbrock(Problem):
    def __init__(self, ndim=1, w=0.2):
        super().__init__()
        self.problem = ReducedProblem(
            Rosenbrock(ndim=ndim + 1), np.arange(1, ndim + 1), w=w
        )

        self.options = OptionsDictionary()
        self.options.declare("ndim", ndim, types=int)
        self.options.declare("return_complex", False, types=bool)
        self.options.declare("name", "NdimRosenbrock", types=str)

        self.xlimits = self.problem.xlimits

    def _evaluate(self, x, kx):
        return self.problem._evaluate(x, kx)



"""
Author: John T. Hwang <hwangjt@umich.edu>
        Mohamed A. Bouhlel <mbouhlel@umich.edu>
        Remi Lafage <remi.lafage@onera.fr>
        Lucas Alber <lucasd.alber@gmail.com>
        Paul Saves <paul.saves@onera.fr>

This package is distributed under New BSD license.
"""

from setuptools import setup, Extension
import sys
import numpy as np
from Cython.Build import cythonize

from smt import __version__

CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: BSD License
Programming Language :: C++
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.8
Programming Language :: Python :: Implementation :: CPython
Topic :: Software Development
Topic :: Scientific/Engineering
Operating System :: Microsoft :: Windows
Operating System :: Unix
Operating System :: MacOS
"""

LONG_DESCRIPTION = """
The surrogate modeling toolbox (SMT) is a Python package that contains \
a collection of surrogate modeling methods, sampling techniques, and \
benchmarking functions. This package provides a library of surrogate \
models that is simple to use and facilitates the implementation of additional methods. \

SMT is different from existing surrogate modeling libraries because of \
its emphasis on derivatives, including training derivatives used for \
gradient-enhanced modeling, prediction derivatives, and derivatives \
with respect to the training data. It also includes new surrogate models \
that are not available elsewhere: kriging by partial-least squares reduction \
and energy-minimizing spline interpolation.

SMT 2.0 adds the capability to handle mixed-variable surrogate models \
and hierarchical variables.
"""

extra_compile_args = []
if not sys.platform.startswith("win"):
    extra_compile_args.append("-std=c++11")

ext = (
    cythonize(
        Extension(
            "smt.surrogate_models.rbfclib",
            sources=["smt/src/rbf/rbf.cpp", "smt/src/rbf/rbfclib.pyx"],
            language="c++",
            extra_compile_args=extra_compile_args,
            include_dirs=[np.get_include()],
        )
    )
    + cythonize(
        Extension(
            "smt.surrogate_models.idwclib",
            sources=["smt/src/idw/idw.cpp", "smt/src/idw/idwclib.pyx"],
            language="c++",
            extra_compile_args=extra_compile_args,
            include_dirs=[np.get_include()],
        )
    )
    + cythonize(
        Extension(
            "smt.surrogate_models.rmtsclib",
            sources=[
                "smt/src/rmts/rmtsclib.pyx",
                "smt/src/rmts/utils.cpp",
                "smt/src/rmts/rmts.cpp",
                "smt/src/rmts/rmtb.cpp",
                "smt/src/rmts/rmtc.cpp",
            ],
            language="c++",
            extra_compile_args=extra_compile_args,
            include_dirs=[np.get_include()],
        )
    )
)


metadata = dict(
    name="smt",
    version=__version__,
    description="The Surrogate Modeling Toolbox (SMT)",
    long_description=LONG_DESCRIPTION,
    author="Remi Lafage et al.",
    author_email="remi.lafage@onera.fr",
    license="BSD-3",
    classifiers=[_f for _f in CLASSIFIERS.split("\n") if _f],
    packages=[
        "smt",
        "smt.surrogate_models",
        "smt.problems",
        "smt.sampling_methods",
        "smt.utils",
        "smt.applications",
    ],
    install_requires=[
        "scikit-learn",
        "pyDOE3",
        "scipy",
        "jenn",
    ],
    extras_require={
        "numba": [  # pip install smt[numba]
            "numba~=0.56.4",
        ],
        "cs": [  # pip install smt[cs]
            "ConfigSpace~=0.6.1",
        ],
        "gpx": ["egobox~=0.19"],  # pip install smt[gpx]
    },
    python_requires=">=3.8",
    zip_safe=False,
    ext_modules=ext,
    url="https://github.com/SMTorg/smt",  # use the URL to the github repo
    download_url="https://github.com/SMTorg/smt/releases",
)

setup(**metadata)
